# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the catalog author module."""

import argparse
import copy
import pathlib
import shutil
import sys

from _pytest.monkeypatch import MonkeyPatch

import pytest

from ruamel.yaml import YAML

from tests import test_utils

from trestle.cli import Trestle
from trestle.common import const, file_utils
from trestle.common.model_utils import ModelUtils
from trestle.core.catalog.catalog_api import CatalogAPI
from trestle.core.catalog.catalog_interface import CatalogInterface
from trestle.core.catalog.catalog_merger import CatalogMerger
from trestle.core.commands.author.catalog import CatalogAssemble, CatalogGenerate
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.commands.import_ import ImportCmd
from trestle.core.control_context import ContextPurpose, ControlContext
from trestle.core.control_interface import ControlInterface, ParameterRep
from trestle.core.markdown.markdown_api import MarkdownAPI
from trestle.core.models.file_content_type import FileContentType
from trestle.core.profile_resolver import ProfileResolver
from trestle.oscal import catalog as cat
from trestle.oscal import profile as prof
from trestle.oscal.common import Part, Property

markdown_name = 'my_md'


def _change_params(ac1_path: pathlib.Path, new_prose: str, make_change: bool) -> None:
    if make_change:
        assert file_utils.insert_text_in_file(ac1_path, 'Procedures {{', f'- \\[d\\] {new_prose}\n')
    assert test_utils.delete_line_in_file(ac1_path, 'ac-1_prm_1', 1)
    assert test_utils.replace_line_in_file_after_tag(
        ac1_path, 'trestle-set-params', '  ac-1_prm_1:\n    values:\n      - new value\n'
    )
    assert test_utils.replace_line_in_file_after_tag(
        ac1_path, 'ac-1_prm_3', '    values:\n      - added param 3 value\n'
    )


@pytest.mark.parametrize('set_parameters_flag', [True, False])
@pytest.mark.parametrize('make_change', [True, False])
@pytest.mark.parametrize('use_orig_cat', [True, False])
@pytest.mark.parametrize('add_header', [True, False])
@pytest.mark.parametrize('use_cli', [True, False])
@pytest.mark.parametrize('dir_exists', [True, False])
def test_catalog_generate_assemble(
    set_parameters_flag: bool,
    make_change: bool,
    use_orig_cat: bool,
    add_header: bool,
    use_cli: bool,
    dir_exists: bool,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch
) -> None:
    """Test the catalog markdown generator."""
    nist_catalog_path = test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME
    cat_name = 'my_cat'
    md_name = 'my_md'
    assembled_cat_name = 'my_assembled_cat'
    catalog_dir = tmp_trestle_dir / f'catalogs/{cat_name}'
    catalog_dir.mkdir(parents=True, exist_ok=True)
    catalog_path = catalog_dir / 'catalog.json'
    shutil.copy(nist_catalog_path, catalog_path)
    markdown_path = tmp_trestle_dir / md_name
    markdown_path.mkdir(parents=True, exist_ok=True)
    ac1_path = markdown_path / 'ac/ac-1.md'
    new_prose = 'My added item'
    assembled_cat_dir = tmp_trestle_dir / f'catalogs/{assembled_cat_name}'
    yaml_header_path = test_utils.YAML_TEST_DATA_PATH / 'good_simple.yaml'

    context = ControlContext.generate(ContextPurpose.CATALOG, True, tmp_trestle_dir, markdown_path)
    context.set_parameters_flag = set_parameters_flag

    # convert catalog to markdown then assemble it after adding an item to a control
    if use_cli:
        test_args = f'trestle author catalog-generate -n {cat_name} -o {md_name}'.split()
        if add_header:
            test_args.extend(['-y', str(yaml_header_path)])
        monkeypatch.setattr(sys, 'argv', test_args)
        assert Trestle().run() == 0
        assert ac1_path.exists()
        _change_params(ac1_path, new_prose, make_change)
        test_args = f'trestle author catalog-assemble -m {md_name} -o {assembled_cat_name}'.split()
        if set_parameters_flag:
            test_args.append('-sp')
        if use_orig_cat:
            test_args.extend(f'-n {cat_name}'.split())
        if dir_exists:
            assembled_cat_dir.mkdir()
        monkeypatch.setattr(sys, 'argv', test_args)
        assert Trestle().run() == 0
    else:
        catalog_generate = CatalogGenerate()
        yaml_header = {}
        if add_header:
            yaml = YAML(typ='safe')
            yaml_header = yaml.load(yaml_header_path.open('r'))
        assert CmdReturnCodes.SUCCESS.value == catalog_generate.generate_markdown(
            tmp_trestle_dir, catalog_path, markdown_path, yaml_header, False
        )
        assert (markdown_path / 'ac/ac-1.md').exists()
        _change_params(ac1_path, new_prose, make_change)
        if dir_exists:
            assembled_cat_dir.mkdir()
        orig_cat_name = cat_name if use_orig_cat else None
        assert CmdReturnCodes.SUCCESS.value == CatalogAssemble.assemble_catalog(
            tmp_trestle_dir, md_name, assembled_cat_name, orig_cat_name, set_parameters_flag, False, ''
        )

    orig_cat: cat.Catalog = cat.Catalog.oscal_read(catalog_path)
    assembled_cat: cat.Catalog = cat.Catalog.oscal_read(assembled_cat_dir / 'catalog.json')
    assert (orig_cat.metadata.title == assembled_cat.metadata.title) == use_orig_cat
    interface_orig = CatalogInterface(orig_cat)
    # need to delete withdrawn controls because they won't be in the assembled catalog
    interface_orig.delete_withdrawn_controls()
    ac1 = interface_orig.get_control('ac-1')
    ac44 = interface_orig.get_control('ac-4.4')
    if make_change:
        # add the item manually to the original catalog so we can confirm the item was loaded correctly
        prop = Property(name='label', value='d.')
        new_part = Part(id='ac-1_smt.d', name='item', props=[prop], prose=new_prose)
        ac1.parts[0].parts.append(new_part)
        interface_orig.replace_control(ac1)
        orig_cat = interface_orig.get_catalog()
    if set_parameters_flag:
        ac1.params[0].values = ['new value']
        ac1.params[2].values = ['added param 3 value']
        interface_orig.replace_control(ac1)
        orig_cat = interface_orig.get_catalog()
    elif not use_orig_cat:
        ac1.params = None
        # ac 4.4 has a parameter set in it that needs to be removed if set_param=False and use_orig_cat=False
        ac44.params = None
        interface_orig.replace_control(ac1)
        orig_cat = interface_orig.get_catalog()
    if use_orig_cat:
        ac1 = assembled_cat.groups[0].controls[0]
        assert ac1.props[2].name == 'extra_prop'
        assert ac1.props[2].value == 'extra value'
        assert ac1.parts[0].props[0].name == 'prop_in_part'
        assert ac1.parts[0].props[0].value == 'value in part'
    assert test_utils.catalog_interface_equivalent(interface_orig, assembled_cat, False)


def test_catalog_assemble_version(sample_catalog_rich_controls: cat.Catalog, tmp_trestle_dir: pathlib.Path) -> None:
    """Test catalog assemble version."""
    cat_name = 'my_cat'
    md_name = 'my_md'
    new_version = '1.2.3'
    assembled_cat_name = 'my_assembled_cat'
    catalog_dir = tmp_trestle_dir / f'catalogs/{cat_name}'
    catalog_dir.mkdir(parents=True, exist_ok=True)
    catalog_path = catalog_dir / 'catalog.json'
    sample_catalog_rich_controls.oscal_write(catalog_path)
    markdown_path = tmp_trestle_dir / md_name
    catalog_generate = CatalogGenerate()

    # generate the catalog markdown
    assert CmdReturnCodes.SUCCESS.value == catalog_generate.generate_markdown(
        tmp_trestle_dir, catalog_path, markdown_path, {}, False
    )

    # assemble the catalog the first time
    assert CmdReturnCodes.SUCCESS.value == CatalogAssemble.assemble_catalog(
        tmp_trestle_dir, md_name, assembled_cat_name, cat_name, False, False, new_version
    )

    # load the freshly assembled catalog
    assembled_cat, assembled_cat_path = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        assembled_cat_name,
        cat.Catalog
    )

    # confirm it is a fresh file with the version set as requested
    assert assembled_cat.metadata.version == new_version
    assert ModelUtils.model_age(assembled_cat) < test_utils.NEW_MODEL_AGE_SECONDS
    creation_time = assembled_cat_path.stat().st_mtime

    # assemble same way again and confirm no new write
    assert CmdReturnCodes.SUCCESS.value == CatalogAssemble.assemble_catalog(
        tmp_trestle_dir, md_name, assembled_cat_name, None, False, False, new_version
    )

    assert creation_time == assembled_cat_path.stat().st_mtime

    # assemble same way again with parent name specified and confirm no new write
    assert CmdReturnCodes.SUCCESS.value == CatalogAssemble.assemble_catalog(
        tmp_trestle_dir, md_name, assembled_cat_name, assembled_cat_name, False, False, new_version
    )

    assert creation_time == assembled_cat_path.stat().st_mtime

    # change version and confirm write
    assert CmdReturnCodes.SUCCESS.value == CatalogAssemble.assemble_catalog(
        tmp_trestle_dir, md_name, assembled_cat_name, assembled_cat_name, False, False, 'xx'
    )

    assert creation_time < assembled_cat_path.stat().st_mtime

    control_text = """# control_q - \[The xy control group\] this is control q

## Control Statement
"""
    # add a new markdown control and make sure it is assembled
    control_path = tmp_trestle_dir / 'my_md/xy/control_q.md'
    with open(control_path, 'w') as f:
        f.write(control_text)
    assert CmdReturnCodes.SUCCESS.value == CatalogAssemble.assemble_catalog(
        tmp_trestle_dir, md_name, assembled_cat_name, assembled_cat_name, False, False, 'xx2'
    )

    catalog, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, 'my_assembled_cat', cat.Catalog, FileContentType.JSON)
    interface = CatalogInterface(catalog)
    assert interface.get_count_of_controls_in_catalog(True) == 7

    # make additions to a sub control and confirm they end up in the assembled catalog
    control_d1_text = """---
x-trestle-set-params:
  param_new:
    values: new param value
---
# control_d1 - \[\] this is control d1

## Control Statement

New control statement.
"""
    control_d1_path = tmp_trestle_dir / 'my_md/control_d1.md'
    with open(control_d1_path, 'w') as f:
        f.write(control_d1_text)
    # need to set parameters during assembly
    assert CmdReturnCodes.SUCCESS.value == CatalogAssemble.assemble_catalog(
        tmp_trestle_dir, md_name, assembled_cat_name, assembled_cat_name, True, False, 'xx3'
    )

    catalog, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, 'my_assembled_cat', cat.Catalog, FileContentType.JSON)
    interface = CatalogInterface(catalog)
    assert interface.get_count_of_controls_in_catalog(True) == 7
    control_d = interface.get_control('control_d')
    assert control_d.controls[0].params[0].id == 'param_new'
    assert control_d.controls[0].params[0].values[0] == 'new param value'


def test_catalog_interface(sample_catalog_rich_controls: cat.Catalog) -> None:
    """Test the catalog interface with complex controls."""
    interface = CatalogInterface(sample_catalog_rich_controls)
    n_controls = interface.get_count_of_controls_in_catalog(True)
    assert n_controls == 6

    control = interface.get_control('control_d1')
    new_title = 'updated d1'
    control.title = new_title
    interface.replace_control(control)
    interface.update_catalog_controls()
    assert interface._catalog.controls[1].controls[0].title == new_title


def test_get_statement_parts(simplified_nist_catalog: cat.Catalog) -> None:
    """Test the catalog interface with complex controls."""
    interface = CatalogInterface(simplified_nist_catalog)
    parts = interface.get_statement_parts('ac-1')
    assert len(parts) == 10
    prt = parts[0]
    assert prt['indent'] == 0
    assert prt['label'] == ''
    assert prt['prose'] == 'The organization:'
    prt = parts[4]
    assert prt['indent'] == 3
    assert prt['label'] == '(b)'
    assert prt['prose'].startswith('Is consistent with')


def test_catalog_interface_groups() -> None:
    """Test handling of groups of groups in CatalogInterface."""
    catalog: cat.Catalog = cat.Catalog.oscal_read(test_utils.JSON_TEST_DATA_PATH / 'nist_tutorial_catalog.json')
    interface = CatalogInterface(catalog)
    interface.update_catalog_controls()
    assert interface.get_count_of_controls_in_catalog(True) == 4
    assert interface.get_count_of_controls_in_catalog(False) == 4
    groups = list(interface.get_all_groups_from_catalog())
    assert len(groups) == 4


def test_get_sorted_controls_in_group(simplified_nist_catalog: cat.Catalog) -> None:
    """Test get sorted controls in group."""
    catalog_interface = CatalogInterface(simplified_nist_catalog)
    controls = catalog_interface.get_sorted_controls_in_group('ac')
    ids = [control.id for control in controls]
    # Confirm the start of list is not alphabetical but is numeric since it uses the control sort-ids for order
    ref_ids = ['ac-1', 'ac-2', 'ac-2.1', 'ac-2.2', 'ac-2.3', 'ac-2.4', 'ac-2.5', 'ac-2.6', 'ac-2.7', 'ac-2.8', 'ac-2.9']
    assert ids[:11] == ref_ids


@pytest.mark.parametrize('replace_params', [True, False])
def test_catalog_interface_merge_controls(replace_params: bool, sample_catalog_rich_controls: cat.Catalog) -> None:
    """Test merging of controls."""
    control_a = sample_catalog_rich_controls.groups[0].controls[0]
    control_b = copy.deepcopy(control_a)
    CatalogMerger.merge_controls(control_a, control_b, replace_params)
    assert control_a == control_b
    control_b.params[0].values = ['new value']
    CatalogMerger.merge_controls(control_a, control_b, replace_params)
    if replace_params:
        assert control_a.params[0].values[0] == 'new value'
    else:
        assert control_a.params[0].values[0] == 'param_0_val'
    control_b.params = control_b.params[:1]
    CatalogMerger.merge_controls(control_a, control_b, replace_params)
    assert len(control_a.params) == 1 if replace_params else 2


def test_catalog_generate_failures(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test failures of author catalog."""
    # disallowed output name
    test_args = 'trestle author catalog-generate -n foo -o profiles'.split()
    monkeypatch.setattr(sys, 'argv', test_args)
    assert Trestle().run() == 1

    # catalog doesn't exist
    test_args = 'trestle author catalog-generate -n foo -o my_md'.split()
    monkeypatch.setattr(sys, 'argv', test_args)
    assert Trestle().run() == 1

    # bad yaml
    bad_yaml_path = str(test_utils.YAML_TEST_DATA_PATH / 'bad_simple.yaml')
    test_args = f'trestle author catalog-generate -n foo -o my_md -y {bad_yaml_path}'.split()
    monkeypatch.setattr(sys, 'argv', test_args)
    assert Trestle().run() == 1


def test_catalog_assemble_failures(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test failurs of catalog assemble."""
    test_args = 'trestle author catalog-assemble -m foo -o my_md'.split()
    monkeypatch.setattr(sys, 'argv', test_args)
    assert Trestle().run() == 1

    (tmp_trestle_dir / 'foo').mkdir()
    monkeypatch.setattr(sys, 'argv', test_args)
    assert Trestle().run() == 1


def test_get_profile_param_dict(tmp_trestle_dir: pathlib.Path) -> None:
    """Test get profile param dict for control."""
    test_utils.setup_for_multi_profile(tmp_trestle_dir, False, True)
    profile, profile_path = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        'test_profile_a',
        prof.Profile,
        FileContentType.JSON
    )
    profile_resolver = ProfileResolver()
    catalog = profile_resolver.get_resolved_profile_catalog(tmp_trestle_dir, profile_path)
    catalog_interface = CatalogInterface(catalog)
    control = catalog_interface.get_control('ac-1')

    full_param_dict = CatalogInterface._get_full_profile_param_dict(profile)
    control_param_dict = CatalogInterface._get_profile_param_dict(control, full_param_dict, False)
    assert ControlInterface.param_to_str(
        control_param_dict['ac-1_prm_1'], ParameterRep.VALUE_OR_LABEL_OR_CHOICES
    ) == 'all alert personnel'
    assert ControlInterface.param_to_str(
        control_param_dict['ac-1_prm_6'], ParameterRep.VALUE_OR_LABEL_OR_CHOICES
    ) == 'monthly'
    # param 7 has no value so its label will be used
    assert ControlInterface.param_to_str(
        control_param_dict['ac-1_prm_7'], ParameterRep.VALUE_OR_LABEL_OR_CHOICES
    ) == 'organization-defined events'


def test_catalog_generate_withdrawn(tmp_path: pathlib.Path, sample_catalog_rich_controls: cat.Catalog) -> None:
    """Test catalog generate when some controls are marked withdrawn."""
    control_a = sample_catalog_rich_controls.groups[0].controls[0]
    control_b = sample_catalog_rich_controls.groups[0].controls[1]
    group_id = sample_catalog_rich_controls.groups[0].id
    if not control_b.props:
        control_b.props = []
    control_b.props.append(Property(name='status', value='Withdrawn'))
    context = ControlContext.generate(ContextPurpose.CATALOG, True, tmp_path, tmp_path)
    catalog_api = CatalogAPI(catalog=sample_catalog_rich_controls, context=context)
    catalog_api.write_catalog_as_markdown()
    # confirm that the first control was written out but not the second
    path_a = tmp_path / group_id / (control_a.id + '.md')
    assert path_a.exists()
    path_b = tmp_path / group_id / (control_b.id + '.md')
    assert not path_b.exists()


def test_params_in_choice(
    tmp_trestle_dir: pathlib.Path, simplified_nist_catalog: cat.Catalog, simplified_nist_profile: prof.Profile
) -> None:
    """Confirm that parameters in choices are substituted properly."""
    # the nist simplified profile defines ac-4.4_prm_3, which is in the choices of ac-4.4_prm_2
    cat_name = 'simplified_nist_catalog'
    prof_name = 'simplified_nist_profile'
    ModelUtils.save_top_level_model(simplified_nist_catalog, tmp_trestle_dir, cat_name, FileContentType.JSON)
    ModelUtils.save_top_level_model(simplified_nist_profile, tmp_trestle_dir, prof_name, FileContentType.JSON)
    prof_path = ModelUtils.get_model_path_for_name_and_class(tmp_trestle_dir, prof_name, prof.Profile)
    catalog = ProfileResolver.get_resolved_profile_catalog(
        tmp_trestle_dir, prof_path, param_rep=ParameterRep.VALUE_OR_LABEL_OR_CHOICES
    )
    cat_interface = CatalogInterface(catalog)
    control = cat_interface.get_control('ac-4.4')
    val_3 = 'hacking the system'
    # confirm the choice text was set properly
    assert control.params[1].select.choice[3] == val_3
    assert control.params[2].values[0] == val_3
    assert catalog.params[1].values[0] == 'loose_2_val_from_prof'

    control = cat_interface.get_control('ac-1')
    param = control.params[0]
    assert param.props[0].value == 'prop value from prof'
    assert param.props[1].value == 'new prop value from prof'
    assert param.links[0].text == 'new text from prof'
    assert param.links[1].text == 'new link text'
    assert param.constraints[1].description == 'new constraint'
    assert param.guidelines[1].prose == 'new guideline'


def test_pulled_params_in_choice(
    tmp_trestle_dir: pathlib.Path, simplified_nist_catalog: cat.Catalog, simplified_nist_profile: prof.Profile
) -> None:
    """Confirm that parameters in choices are substituted properly and give lower priority to upstream subs."""
    # the nist catalog defines ac-4.4_prm_3, which is referenced by the choices of ac-4.4_prm_2
    # the nist simplified profile itself defines ac-4.4_prm_3 and thereby defines the choices in ac-4.4_prm_2
    # but it is also set by the pulling profile, which should have final say
    # in addition, both the simplified profile and the pulling profile define the desired values for ac-4.4_prm_2
    # the provided values need to be a subset of the choice options
    # the pulling profile should dictate what happens in a generate-assemble cycle for values and choices
    # so the generated markdown needs to substitute current set-param values into the choices instead of holding back
    cat_name = 'simplified_nist_catalog'
    prof_name = 'simplified_nist_profile'
    ModelUtils.save_top_level_model(simplified_nist_catalog, tmp_trestle_dir, cat_name, FileContentType.JSON)
    ModelUtils.save_top_level_model(simplified_nist_profile, tmp_trestle_dir, prof_name, FileContentType.JSON)
    pull_prof_name = 'pull_nist_profile'
    prof_path = ModelUtils.get_model_path_for_name_and_class(
        tmp_trestle_dir, pull_prof_name, prof.Profile, FileContentType.JSON
    )
    prof_path.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(test_utils.JSON_TEST_DATA_PATH / (pull_prof_name + '.json'), prof_path)

    # get the resolved profile catalog
    catalog = ProfileResolver.get_resolved_profile_catalog(
        tmp_trestle_dir, prof_path, param_rep=ParameterRep.VALUE_OR_LABEL_OR_CHOICES
    )
    cat_interface = CatalogInterface(catalog)

    # check values
    control = cat_interface.get_control('ac-4.4')
    val_1 = 'blocking the flow of the encrypted information'
    val_2 = 'from pulling profile'
    assert control.params[1].values[0] == val_1
    assert control.params[1].values[1] == val_2

    # confirm the choice text was handled properly
    # the param value and the choice should be set by the pulling profile
    assert control.params[2].values[0] == val_2
    assert control.params[1].select.choice[3] == val_2

    # this confirms the pulling profile sets the value of a loose param
    assert catalog.params[1].values[0] == 'loose_2_val_from_pulling'

    # confirm other attributes are as expected
    control = cat_interface.get_control('ac-1')
    param = control.params[0]
    assert param.props[0].value == 'prop value from prof'
    assert param.props[1].value == 'new prop value from prof'
    assert param.links[0].text == 'new text from prof'
    assert param.links[1].text == 'new link text'
    assert param.constraints[1].description == 'new constraint'
    assert param.guidelines[1].prose == 'new guideline'


def test_generate_group_id() -> None:
    """Test the generation of group ids."""
    cat_int = CatalogInterface()
    group = cat.Group(title='my test title')
    assert cat_int._generate_group_id(group) == 'trestle_group_0000'
    assert cat_int._generate_group_id(group) == 'trestle_group_0001'


def test_validate_catalog_missing_group_id(
    tmp_trestle_dir: pathlib.Path, tmp_path_factory: pytest.TempPathFactory, sample_catalog_rich_controls: cat.Catalog
) -> None:
    """Test validation of catalog with groups that dont have ids."""
    # kill one of the group id's
    sample_catalog_rich_controls.groups[0].id = None
    cat_name = 'my_cat'
    tmp_cat_path = tmp_path_factory.mktemp('temp_dir') / (cat_name + '.json')

    # write catalog with missing group id to tmp location
    sample_catalog_rich_controls.oscal_write(tmp_cat_path)
    import_cmd = ImportCmd()
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, file=str(tmp_cat_path), output=cat_name, verbose=1, regenerate=False
    )

    # import the file with missing group id into the trestle workspace
    # this should validate the file and assign the missing id a new value in memory - then save the file
    rc = import_cmd._run(args)
    assert rc == 0

    # load the file without doing validation - to make sure the file itself has the group id assigned
    new_cat, new_cat_path = ModelUtils.load_model_for_class(tmp_trestle_dir, cat_name, cat.Catalog)
    assert new_cat.groups[0].id == 'trestle_group_0000'

    md_name = 'md_cat'
    md_path = tmp_trestle_dir / md_name
    md_path.mkdir(parents=True, exist_ok=True)
    cat_generate = CatalogGenerate()
    assert cat_generate.generate_markdown(
        tmp_trestle_dir, new_cat_path, md_path, {}, False
    ) == CmdReturnCodes.SUCCESS.value

    assem_cat_name = 'assem_cat'
    cat_assemble = CatalogAssemble()
    cat_assemble.assemble_catalog(tmp_trestle_dir, md_name, assem_cat_name, None, False, False, None)

    # load the file without doing validation - to make sure the file itself has the group id assigned
    _, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, assem_cat_name, cat.Catalog)
    assert new_cat.groups[0].id == 'trestle_group_0000'


def test_get_control_paths(sample_catalog_rich_controls: cat.Catalog) -> None:
    """Test get control paths."""
    cat_interface = CatalogInterface(sample_catalog_rich_controls)
    path = cat_interface.get_full_control_path('control_s')
    assert path == ['xy', 'sub']
    assert cat_interface.get_control_path('control_s') == ['xy', 'sub']
    path = cat_interface.get_full_control_path('control_d1')
    assert path == ['', 'control_d']
    assert cat_interface.get_control_path('control_d1') == ['']
    control = copy.deepcopy(cat_interface.get_control('control_d1'))
    control.id = 'cat_level'
    sample_catalog_rich_controls.controls = [control]
    cat_interface = CatalogInterface(sample_catalog_rich_controls)
    path = cat_interface.get_full_control_path('cat_level')
    assert path == ['']
    assert cat_interface.get_control_path('cat_level') == ['']


def test_catalog_force_overwrite(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test catalog generate with force-overwrite."""
    catalog = cat.Catalog.oscal_read(test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME)
    ModelUtils.save_top_level_model(catalog, tmp_trestle_dir, 'my_catalog', FileContentType.JSON)

    catalog_generate = 'trestle author catalog-generate -n my_catalog -o md_catalog --force-overwrite'
    test_utils.execute_command_and_assert(catalog_generate, 0, monkeypatch)

    catalog_generate = 'trestle author catalog-generate -n my_catalog -o md_catalog'
    test_utils.execute_command_and_assert(catalog_generate, 0, monkeypatch)

    md_path = tmp_trestle_dir / 'md_catalog/ac/ac-2.md'
    assert md_path.exists()
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(md_path)

    assert header
    old_value = header[const.SET_PARAMS_TAG]['ac-2_prm_1'][const.VALUES]
    header[const.SET_PARAMS_TAG]['ac-2_prm_1'][const.VALUES] = 'New value'
    md_api.write_markdown_with_header(md_path, header, tree.content.raw_text)

    catalog_generate = 'trestle author catalog-generate -n my_catalog -o md_catalog'
    test_utils.execute_command_and_assert(catalog_generate, 0, monkeypatch)

    header, _ = md_api.processor.process_markdown(md_path)
    assert header[const.SET_PARAMS_TAG]['ac-2_prm_1'][const.VALUES] == 'New value'

    catalog_generate = 'trestle author catalog-generate -n my_catalog -o md_catalog --force-overwrite'
    test_utils.execute_command_and_assert(catalog_generate, 0, monkeypatch)

    header, _ = md_api.processor.process_markdown(md_path)
    assert header[const.SET_PARAMS_TAG]['ac-2_prm_1'][const.VALUES] == old_value

    fc = test_utils.FileChecker(tmp_trestle_dir / 'md_catalog/')
    catalog_generate = 'trestle author catalog-generate -n my_catalog -o md_catalog --force-overwrite'
    test_utils.execute_command_and_assert(catalog_generate, 0, monkeypatch)
    assert fc.files_unchanged()


def test_prune_written_controls(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test pruning of written controls."""
    catalog = cat.Catalog.oscal_read(test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME)
    ModelUtils.save_top_level_model(catalog, tmp_trestle_dir, 'my_catalog', FileContentType.JSON)

    catalog_generate = 'trestle author catalog-generate -n my_catalog -o md_catalog'
    test_utils.execute_command_and_assert(catalog_generate, 0, monkeypatch)

    md_path = tmp_trestle_dir / 'md_catalog'

    catalog_interface = CatalogInterface(catalog)
    control_ids = set(catalog_interface.get_control_ids())
    controls_to_delete = sorted(['ac-1', 'ac-2.9', 'at-2.1'])
    id_subset = control_ids - set(controls_to_delete)

    assert CatalogInterface._prune_controls(md_path, id_subset) == controls_to_delete


def test_catalog_assemble_subgroups(
    tmp_trestle_dir: pathlib.Path, sample_catalog_subgroups: cat.Catalog, monkeypatch: MonkeyPatch
) -> None:
    """Test assembly of catalog with group having no controls but does contain subgroup."""
    ModelUtils.save_top_level_model(sample_catalog_subgroups, tmp_trestle_dir, 'my_catalog', FileContentType.JSON)
    catalog_generate = 'trestle author catalog-generate -n my_catalog -o md_catalog -vv'
    test_utils.execute_command_and_assert(catalog_generate, 0, monkeypatch)
    catalog_assemble = 'trestle author catalog-assemble -m md_catalog -o my_catalog -vv'
    test_utils.execute_command_and_assert(catalog_assemble, 0, monkeypatch)


def test_catalog_multiline_statement(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test catalog-assemble with multiline prose for control statement."""
    catalog = cat.Catalog.oscal_read(test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME)
    ModelUtils.save_top_level_model(catalog, tmp_trestle_dir, 'my_catalog', FileContentType.JSON)

    catalog_generate = 'trestle author catalog-generate -n my_catalog -o md_catalog'
    test_utils.execute_command_and_assert(catalog_generate, 0, monkeypatch)
    md_path = tmp_trestle_dir / 'md_catalog/ac/ac-2.md'
    assert md_path.exists()

    control_statement_prose = """The organization:

Test 1

Here goes a long paragraph. Test 2

Test 3
Test 4


"""

    file_utils.insert_text_in_file(md_path, '## Control Statement', control_statement_prose)

    catalog_assemble = 'trestle author catalog-assemble -o my_catalog -m md_catalog'
    test_utils.execute_command_and_assert(catalog_assemble, 0, monkeypatch)

    catalog, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, 'my_catalog', cat.Catalog)

    assert catalog
    assert catalog.groups[0].controls[1].parts[0].prose == control_statement_prose.strip('\n')
    assert catalog.groups[0].controls[1].parts[0].parts[
        0
    ].prose == 'Define and document the types of accounts allowed and specifically prohibited for use within the system;'  # noqa E501


def test_catalog_duplicate_parts_statement(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch, capsys) -> None:
    """Test catalog-assemble with duplicate parts for control statement."""
    catalog = cat.Catalog.oscal_read(test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME)
    ModelUtils.save_top_level_model(catalog, tmp_trestle_dir, 'my_catalog', FileContentType.JSON)

    catalog_generate = 'trestle author catalog-generate -n my_catalog -o md_catalog'
    test_utils.execute_command_and_assert(catalog_generate, 0, monkeypatch)
    md_path = tmp_trestle_dir / 'md_catalog/ac/ac-2.md'
    assert md_path.exists()

    control_statement_prose_with_parts = """The organization:

- \[a\] Part A
- \[a\] Part A Duplicate.
  - \[a.1\] Documents 1
  - \[a.2\] Documents 2
    - \[a.2.1\] SubDocuments 1
    - \[a.2.2\] SubDocuments 2
- \[b\] Part 2.


"""

    file_utils.insert_text_in_file(md_path, '## Control Statement', control_statement_prose_with_parts)

    catalog_assemble = 'trestle author catalog-assemble -o my_catalog -m md_catalog'
    test_utils.execute_command_and_assert(catalog_assemble, 0, monkeypatch)

    _, error = capsys.readouterr()
    assert 'Duplicate part id ac-2_smt.a' in error


def test_catalog_tab_in_statement(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch, capsys) -> None:
    """Test catalog-assemble with tab characters in control statement and parts."""
    catalog = cat.Catalog.oscal_read(test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME)
    ModelUtils.save_top_level_model(catalog, tmp_trestle_dir, 'my_catalog', FileContentType.JSON)

    catalog_generate = 'trestle author catalog-generate -n my_catalog -o md_catalog'
    test_utils.execute_command_and_assert(catalog_generate, 0, monkeypatch)
    md_path = tmp_trestle_dir / 'md_catalog/ac/ac-2.md'
    assert md_path.exists()

    control_statement_prose_with_parts = """The organization:
\t
\t
\t
- \[m\] Part M
- \[n\] Part N
\t- \[n.1\] Documents 1
\t- \[n.2\] Documents 2
\t\t- \[n.2.1\] SubDocuments 1
\t\t- \[n.2.2\] SubDocuments 2
- \[o\] Part O.
\t
\t
\t
"""

    file_utils.insert_text_in_file(md_path, '## Control Statement', control_statement_prose_with_parts)

    catalog_assemble = 'trestle author catalog-assemble -o my_catalog -m md_catalog'
    test_utils.execute_command_and_assert(catalog_assemble, 0, monkeypatch)

    catalog, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, 'my_catalog', cat.Catalog)
    statement_part_n = catalog.groups[0].controls[1].parts[0].parts[1]
    assert statement_part_n.id == 'ac-2_smt.n'
    assert statement_part_n.parts[0].id == 'ac-2_smt.n.n.1'



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common fixtures."""
import os
import pathlib
import random
import string
import sys
from typing import Iterator
from uuid import uuid4

from _pytest.monkeypatch import MonkeyPatch

import pytest

from tests import test_utils

import trestle.core.generators as gens
import trestle.oscal.common as common
from trestle.cli import Trestle
from trestle.common.err import TrestleError
from trestle.oscal import catalog as cat
from trestle.oscal.component import ComponentDefinition, DefinedComponent
from trestle.oscal.profile import Profile

TEST_CONFIG: dict = {}


@pytest.fixture(scope='function')
def rand_str():
    """Return a random string."""
    rand_str = ''.join(random.choice(string.ascii_letters) for x in range(16))
    return rand_str


@pytest.fixture(scope='function')
def tmp_file(tmp_path):
    """Return the path for a tmp file."""
    return pathlib.Path(tmp_path) / f'{uuid4()}'


@pytest.fixture(scope='session')
def tmp_fixed_file(tmp_path):
    """Return the path for a tmp file with fixed name."""
    return pathlib.Path(tmp_path) / 'fixed_file'


@pytest.fixture(scope='function')
def tmp_yaml_file(tmp_path):
    """Return the path for a tmp yaml file."""
    return pathlib.Path(tmp_path) / f'{uuid4()}.yaml'


@pytest.fixture(scope='function')
def tmp_json_file(tmp_path):
    """Return the path for a tmp json file."""
    return pathlib.Path(tmp_path) / f'{uuid4()}.json'


@pytest.fixture(scope='function')
def tmp_xml_file(tmp_path):
    """Return the path for a tmp xml file."""
    return pathlib.Path(tmp_path) / f'{uuid4()}.xml'


@pytest.fixture(scope='module')
def yaml_testdata_path() -> pathlib.Path:
    """Return the path for the yaml test data directory."""
    return pathlib.Path(test_utils.YAML_TEST_DATA_PATH)


@pytest.fixture(scope='module')
def json_testdata_path() -> pathlib.Path:
    """Return the path for json test data directory."""
    return pathlib.Path(test_utils.JSON_TEST_DATA_PATH)


@pytest.fixture(scope='function')
def sample_nist_component_def() -> ComponentDefinition:
    """Return a rich component definition object, from the NIST content repository."""
    component_obj = ComponentDefinition.oscal_read(test_utils.NIST_SAMPLE_CD_JSON)
    return component_obj


@pytest.fixture(scope='function')
def sample_catalog():
    """Return a large nist catalog."""
    file_path = pathlib.Path(test_utils.JSON_NIST_DATA_PATH) / test_utils.JSON_NIST_CATALOG_NAME
    catalog_obj = cat.Catalog.oscal_read(file_path)
    return catalog_obj


@pytest.fixture(scope='function')
def sample_profile():
    """Return a large nist profile."""
    file_path = pathlib.Path(test_utils.JSON_NIST_DATA_PATH) / test_utils.JSON_NIST_PROFILE_NAME
    profile_obj = Profile.oscal_read(file_path)
    return profile_obj


@pytest.fixture(scope='function')
def sample_trestle_profile():
    """Return a hand curated profile for testing."""
    file_path = pathlib.Path(test_utils.JSON_TEST_DATA_PATH) / 'simple_test_profile.json'
    profile_obj = Profile.oscal_read(file_path)
    return profile_obj


@pytest.fixture(scope='function')
def simplified_nist_catalog():
    """Return a simplified nist catalog."""
    file_path = pathlib.Path(test_utils.JSON_TEST_DATA_PATH) / test_utils.SIMPLIFIED_NIST_CATALOG_NAME
    catalog_obj = cat.Catalog.oscal_read(file_path)
    return catalog_obj


@pytest.fixture(scope='function')
def simplified_nist_profile():
    """Return a simplified nist profile."""
    file_path = pathlib.Path(test_utils.JSON_TEST_DATA_PATH) / test_utils.SIMPLIFIED_NIST_PROFILE_NAME
    profile_obj = Profile.oscal_read(file_path)
    return profile_obj


@pytest.fixture(scope='function')
def sample_catalog_minimal():
    """Return a valid catalog object with minimum fields necessary."""
    file_path = pathlib.Path(test_utils.JSON_TEST_DATA_PATH) / 'minimal_catalog.json'
    catalog_obj = cat.Catalog.oscal_read(file_path)
    return catalog_obj


@pytest.fixture(scope='function')
def sample_catalog_missing_roles():
    """Return a catalog object missing roles."""
    file_path = pathlib.Path(test_utils.JSON_TEST_DATA_PATH) / 'minimal_catalog_missing_roles.json'
    catalog_obj = cat.Catalog.oscal_read(file_path)
    return catalog_obj


@pytest.fixture(scope='function')
def sample_catalog_rich_controls():
    """Return a catalog with controls in groups and in the catalog itself."""
    catalog_obj = gens.generate_sample_model(cat.Catalog)

    param_0 = common.Parameter(id='param_0', values=['param_0_val'])
    param_1 = common.Parameter(id='param_1', values=['param_1_val'])
    control_a = cat.Control(id='control_a', title='this is control a', params=[param_0, param_1])
    control_b = cat.Control(id='control_b', title='this is control b')
    group = cat.Group(id='xy', title='The xy control group', controls=[control_a, control_b])
    control_s = cat.Control(id='control_s', title='this is control s')
    sub_group = cat.Group(id='sub', title='The sub group', controls=[control_s])
    group.groups = [sub_group]
    catalog_obj.groups = [group]

    part = common.Part(id='cpart', name='name.c.part')
    control_c = cat.Control(id='control_c', title='this is control c', parts=[part])

    control_d = cat.Control(id='control_d', title='this is control d')
    control_d1 = cat.Control(id='control_d1', title='this is control d1')
    control_d.controls = [control_d1]

    catalog_obj.controls = [control_c, control_d]
    return catalog_obj


@pytest.fixture(scope='function')
def sample_catalog_subgroups():
    """Return a catalog with groups of groups."""
    catalog_obj = gens.generate_sample_model(cat.Catalog)
    group = cat.Group(id='a', title='The a group')
    control_c = cat.Control(id='control_c', title='this is control c')
    sub_sub_group = cat.Group(id='bb', title='The bb sub sub group', controls=[control_c])
    sub_group = cat.Group(id='b', title='The b sub group', groups=[sub_sub_group])
    group.groups = [sub_group]
    catalog_obj.groups = [group]

    return catalog_obj


@pytest.fixture(scope='function')
def sample_component_definition():
    """Return a valid ComponentDefinition object with some contents."""
    # one component has no properties - the other has two
    def_comp1: DefinedComponent = gens.generate_sample_model(DefinedComponent, True, 3)
    def_comp2: DefinedComponent = gens.generate_sample_model(DefinedComponent, True, 3)
    prop_1 = common.Property(name='prop_1', value='prop_1_value')
    prop_2 = common.Property(name='prop_2', value='prop_2_value')
    def_comp2.props = [prop_1, prop_2]
    comp_def: ComponentDefinition = gens.generate_sample_model(ComponentDefinition)
    comp_def.components = [def_comp1, def_comp2]
    return comp_def


@pytest.fixture(scope='function')
def tmp_trestle_dir(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch) -> Iterator[pathlib.Path]:
    """Create and return a new trestle workspace directory using std tmp_path fixture.

    Note that this fixture relies on the 'trestle init' command and therefore may
    misbehave if there are errors in trestle init, perhaps in spite of the try block.
    """
    pytest_cwd = pathlib.Path.cwd()
    os.chdir(tmp_path)
    testargs = ['trestle', 'init']
    monkeypatch.setattr(sys, 'argv', testargs)
    try:
        Trestle().run()
    except Exception as e:
        raise TrestleError(f'Initialization failed for temporary trestle directory: {e}.')
    else:
        yield tmp_path
    finally:
        os.chdir(pytest_cwd)


@pytest.fixture(scope='function')
def tmp_empty_cwd(tmp_path: pathlib.Path) -> Iterator[pathlib.Path]:
    """Create a temporary directory and cd into that directory with fail out afterwards.

    The purpose of this is to provide a clean directory per unit test and ensure we get
    back to the base time.
    """
    pytest_cwd = pathlib.Path.cwd()
    os.chdir(tmp_path)

    yield tmp_path

    os.chdir(pytest_cwd)


@pytest.fixture(scope='function')
def testdata_dir() -> pathlib.Path:
    """Return absolute path to test data directory."""
    test_data_source = pathlib.Path('tests/data')
    return test_data_source.resolve()


@pytest.fixture(scope='function')
def keep_cwd() -> Iterator[pathlib.Path]:
    """Force test to return to orig directory if chdir's happen in test."""
    old_cwd = os.getcwd()
    yield old_cwd
    os.chdir(old_cwd)



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OSCAL Profile to OSCO task tests."""

import configparser
import json
import os
import pathlib

from ruamel.yaml import YAML

from tests.test_utils import text_files_equal

import trestle.tasks.oscal_profile_to_osco_profile as oscal_profile_to_osco_profile
from trestle.oscal.profile import Profile
from trestle.tasks.base_task import TaskOutcome
from trestle.transforms.implementations.osco import OscalProfileToOscoProfileTransformer


def setup_config(path: str):
    """Config."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path(path)
    config.read(config_path)
    return config


def test_oscal_profile_to_osco_profile_print_info(tmp_path):
    """Test print_info call."""
    config = setup_config('tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile.config')
    section = config['task.oscal-profile-to-osco-profile']
    section['output-dir'] = str(tmp_path)
    tgt = oscal_profile_to_osco_profile.ProfileToOsco(section)
    retval = tgt.print_info()
    assert retval is None


def test_oscal_profile_to_osco_profile_simulate(tmp_path):
    """Test simulate call."""
    config = setup_config('tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile.config')
    section = config['task.oscal-profile-to-osco-profile']
    section['output-dir'] = str(tmp_path)
    tgt = oscal_profile_to_osco_profile.ProfileToOsco(section)
    retval = tgt.simulate()
    assert retval == TaskOutcome.SIM_SUCCESS
    assert len(os.listdir(str(tmp_path))) == 0


def test_oscal_profile_to_osco_profile_execute(tmp_path):
    """Test execute call."""
    config = setup_config('tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile.config')
    _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config)


def test_oscal_profile_to_osco_profile_execute_osco_0_1_39_parms_no(tmp_path):
    """Test execute call."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-0.1.39-parms-no.config'
    )
    _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config)


def test_oscal_profile_to_osco_profile_execute_osco_0_1_39_parms_yes(tmp_path):
    """Test execute call."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-0.1.39-parms-yes.config'
    )
    _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config)


def test_oscal_profile_to_osco_profile_execute_osco_0_1_40_parms_no(tmp_path):
    """Test execute call."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-0.1.40-parms-no.config'
    )
    _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config)


def test_oscal_profile_to_osco_profile_execute_osco_0_1_40_parms_yes(tmp_path):
    """Test execute call."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-0.1.40-parms-yes.config'
    )
    _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config)


def test_oscal_profile_to_osco_profile_execute_osco_0_2_0_parms_no(tmp_path):
    """Test execute call."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-0.2.0-parms-no.config'
    )
    _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config)


def test_oscal_profile_to_osco_profile_execute_osco_1_0_0_parms_yes(tmp_path):
    """Test execute call."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-1.0.0-parms-yes.config'
    )
    _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config)


def test_oscal_profile_to_osco_profile_execute_osco_parms_no(tmp_path):
    """Test execute call."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-parms-no.config'
    )
    _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config)


def test_oscal_profile_to_osco_profile_execute_osco_parms_yes(tmp_path):
    """Test execute call."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-parms-yes.config'
    )
    _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config)


def test_oscal_profile_to_osco_profile_execute_osco_scc(tmp_path):
    """Test execute call."""
    config = setup_config('tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-scc.config')
    _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config)


def _test_oscal_profile_to_osco_profile_execute_common(tmp_path, config):
    """Test execute call."""
    section = config['task.oscal-profile-to-osco-profile']
    d_expected = pathlib.Path(section['output-dir'])
    d_produced = tmp_path
    section['output-dir'] = str(d_produced)
    tgt = oscal_profile_to_osco_profile.ProfileToOsco(section)
    retval = tgt.simulate()
    assert retval == TaskOutcome.SIM_SUCCESS
    assert len(os.listdir(str(tmp_path))) == 0
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    list_dir = os.listdir(d_produced)
    assert len(list_dir) == 1
    assert d_expected != d_produced
    for fn in list_dir:
        f_expected = d_expected / fn
        f_produced = d_produced / fn
        result = text_files_equal(f_expected, f_produced)
        assert result


def test_oscal_profile_to_osco_profile_execute_bogus_profile(tmp_path):
    """Test execute call bogus profile."""
    config = setup_config('tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-bogus.config')
    section = config['task.oscal-profile-to-osco-profile']
    section['output-dir'] = str(tmp_path)
    tgt = oscal_profile_to_osco_profile.ProfileToOsco(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_oscal_profile_to_osco_profile_execute_bogus_config(tmp_path):
    """Test execute call bogus config."""
    section = None
    tgt = oscal_profile_to_osco_profile.ProfileToOsco(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_oscal_profile_to_osco_profile_execute_no_input_file(tmp_path):
    """Test execute call no input file."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-no-input-file.config'
    )
    section = config['task.oscal-profile-to-osco-profile']
    section['output-dir'] = str(tmp_path)
    tgt = oscal_profile_to_osco_profile.ProfileToOsco(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_oscal_profile_to_osco_profile_execute_no_output_dir(tmp_path):
    """Test execute call no output file."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-no-output-dir.config'
    )
    section = config['task.oscal-profile-to-osco-profile']
    tgt = oscal_profile_to_osco_profile.ProfileToOsco(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_oscal_profile_to_osco_profile_execute_no_overwrite(tmp_path):
    """Test execute call no overwrite."""
    config = setup_config(
        'tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-no-overwrite.config'
    )
    section = config['task.oscal-profile-to-osco-profile']
    d_expected = pathlib.Path(section['output-dir'])
    d_produced = tmp_path
    section['output-dir'] = str(d_produced)
    tgt = oscal_profile_to_osco_profile.ProfileToOsco(section)
    retval = tgt.simulate()
    assert retval == TaskOutcome.SIM_SUCCESS
    assert len(os.listdir(str(tmp_path))) == 0
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    list_dir = os.listdir(d_produced)
    assert len(list_dir) == 1
    assert d_expected != d_produced
    for fn in list_dir:
        f_expected = d_expected / fn
        f_produced = d_produced / fn
        result = text_files_equal(f_expected, f_produced)
        assert result
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_oscal_profile_to_osco_profile_execute_set(tmp_path):
    """Test execute call with set variables."""
    config = setup_config('tests/data/tasks/oscal-profile-to-osco-profile/oscal-profile-to-osco-profile-set.config')
    section = config['task.oscal-profile-to-osco-profile']
    input_file = section['input-file']
    input_path = pathlib.Path(input_file)
    f_expected = pathlib.Path(section['output-dir'], 'osco-profile.yaml')
    d_produced = tmp_path
    section['output-dir'] = str(d_produced)
    f_produced = pathlib.Path(d_produced, 'osco-profile.yaml')
    # read input
    profile = Profile.oscal_read(input_path)
    # transform
    transformer = OscalProfileToOscoProfileTransformer(
        extends='extends', api_version='api_version', kind='kind', name='name', namespace='namespace'
    )
    ydata = json.loads(transformer.transform(profile))
    # write output
    yaml = YAML(typ='safe')
    yaml.default_flow_style = False
    with open(f_produced, 'w') as outfile:
        yaml.dump(ydata, outfile)
    assert f_expected != f_produced
    result = text_files_equal(f_expected, f_produced)
    assert result



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""csv-to-oscal-cd task tests."""

import configparser
import csv
import os
import pathlib
from typing import List
from unittest import mock

from _pytest.monkeypatch import MonkeyPatch

from tests import test_utils

import trestle.tasks.csv_to_oscal_cd as csv_to_oscal_cd
from trestle.oscal.component import ComponentDefinition
from trestle.tasks.base_task import TaskOutcome


def monkey_exception() -> None:
    """Monkey exception."""
    raise RuntimeError('foobar')


def _get_rows(file_: str) -> List[List[str]]:
    """Get rows from csv file."""
    rows = []
    csv_path = pathlib.Path(file_)
    with open(csv_path, 'r', newline='') as f:
        csv_reader = csv.reader(f, delimiter=',', quoting=csv.QUOTE_MINIMAL)
        for row in csv_reader:
            rows.append(row)
    return rows


def _validate_ocp(tmp_path: pathlib.Path) -> None:
    """Validate ocp."""
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    assert cd.metadata.title == 'Component definition for CIS Red Hat OpenShift Container Platform 4 Benchmark profiles'
    assert cd.metadata.version == 'V1.1'
    assert len(cd.components) == 1
    component = cd.components[0]
    assert len(component.props) == 429
    assert component.type == 'Service'
    assert component.title == 'OSCO'
    assert component.props[0].name == 'Rule_Id'
    assert component.props[0].ns == 'http://oscal-compass.github.io/compliance-trestle/schemas/oscal/cd'
    assert component.props[0].value == 'xccdf_org.ssgproject.content_rule_api_server_anonymous_auth'
    assert component.props[0].class_ == 'scc_class'
    assert component.props[0].remarks == 'rule_set_000'
    assert component.props[1].name == 'Rule_Description'
    assert component.props[1].value == 'Ensure that the --anonymous-auth argument is set to false'
    assert component.props[1].remarks == 'rule_set_000'
    assert component.props[2].name == 'Parameter_Id'
    assert component.props[2].value == 'scan_interval_max'
    assert component.props[3].name == 'Parameter_Description'
    assert component.props[3].value == 'Max Scan Interval Days'
    assert component.props[4].name == 'Parameter_Value_Alternatives'
    assert component.props[4].value == '10, 30'
    assert component.props[5].name == 'Private_Reference_Id'
    assert component.props[5].value == '300000100'
    assert component.props[6].name == 'Rule_Id'
    assert component.props[6].value == 'xccdf_org.ssgproject.content_rule_api_server_basic_auth'
    assert component.props[6].remarks == 'rule_set_001'
    assert component.props[428].name == 'Rule_Description'
    assert component.props[428].value == 'Ensure that the --protect-kernel-defaults argument is set to true'
    assert component.props[428].remarks == 'rule_set_200'
    assert component.props[0].name == 'Rule_Id'
    assert len(component.control_implementations) == 2
    assert component.control_implementations[0].description == 'ocp4'
    assert len(component.control_implementations[0].implemented_requirements) == 88
    assert component.control_implementations[0].implemented_requirements[0].control_id == 'CIS-1.2.1'
    assert component.control_implementations[0].implemented_requirements[0].props[0].name == 'Rule_Id'
    assert component.control_implementations[0].implemented_requirements[0].props[
        0].value == 'xccdf_org.ssgproject.content_rule_api_server_anonymous_auth'
    assert component.control_implementations[1].description == 'ocp4-node'
    assert len(component.control_implementations[1].implemented_requirements) == 31
    assert component.control_implementations[1].implemented_requirements[0].control_id == 'CIS-1.1.1'
    assert component.control_implementations[1].implemented_requirements[30].props[
        0].value == 'xccdf_org.ssgproject.content_rule_kubelet_enable_protect_kernel_sysctl'
    assert component.control_implementations[1].implemented_requirements[30].props[
        1].value == 'xccdf_org.ssgproject.content_rule_kubelet_enable_protect_kernel_defaults'


def _validate_bp(tmp_path: pathlib.Path) -> None:
    """Validate bp."""
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    assert cd.metadata.title == 'Component definition for Best Practices'
    assert cd.metadata.version == 'V1.1'
    assert len(cd.components) == 1
    component = cd.components[0]
    assert len(component.props) == 62
    assert component.type == 'Service'
    assert component.title == 'IAM'
    assert len(component.control_implementations) == 1
    assert component.control_implementations[
        0].description == 'NIST Special Publication 800-53 Revision 5 HIGH IMPACT BASELINE'


def _test_init(tmp_path: pathlib.Path):
    """Test init."""
    test_utils.ensure_trestle_config_dir(tmp_path)


def _get_config_section(tmp_path: pathlib.Path, fname: str) -> tuple:
    """Get config section."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path(f'tests/data/tasks/csv-to-oscal-cd/{fname}')
    config.read(config_path)
    section = config['task.csv-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    return (config, section)


def _get_config_section_init(tmp_path: pathlib.Path, fname: str) -> tuple:
    """Get config section."""
    _test_init(tmp_path)
    return _get_config_section(tmp_path, fname)


def test_print_info(tmp_path: pathlib.Path) -> None:
    """Test print_info."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.print_info()
    assert retval is None


def test_simulate(tmp_path: pathlib.Path) -> None:
    """Test simulate."""
    _, section = _get_config_section(tmp_path, 'test-csv-to-oscal-cd.config')
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.simulate()
    assert retval == TaskOutcome.SIM_SUCCESS
    assert len(os.listdir(str(tmp_path))) == 0


def test_execute(tmp_path: pathlib.Path) -> None:
    """Test execute."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_ocp(tmp_path)


def test_config_missing(tmp_path: pathlib.Path) -> None:
    """Test config missing."""
    section = None
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_config_missing_title(tmp_path: pathlib.Path) -> None:
    """Test config missing title."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    section.pop('title')
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_config_missing_version(tmp_path: pathlib.Path) -> None:
    """Test config missing version."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    section.pop('version')
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_config_missing_csv_file_spec(tmp_path: pathlib.Path) -> None:
    """Test config missing csv file spec."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    section['output-dir'] = str(tmp_path)
    section.pop('csv-file')
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_config_missing_csv_file(tmp_path: pathlib.Path) -> None:
    """Test config missing csv file."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    section['csv-file'] = 'foobar'
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_exception(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test exception."""
    monkeypatch.setattr(csv_to_oscal_cd._RuleSetIdMgr, 'get_next_rule_set_id', monkey_exception)
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_execute_mock(tmp_path: pathlib.Path) -> None:
    """Test execute mock."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    # get good data & test that mocking works
    rows = _get_rows('tests/data/csv/ocp4-user.v2.csv')
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
        _validate_ocp(tmp_path)


def _setup_workspace(workspace: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test workspace setup."""
    src = 'tests/data/tasks/csv-to-oscal-cd'
    tgt = workspace / src
    test_utils.copy_tree_with_hidden(src, tgt)
    src = 'tests/data/csv'
    tgt = workspace / src
    test_utils.copy_tree_with_hidden(src, tgt)
    src = 'tests/data/ocp4-cis'
    tgt = workspace / src
    test_utils.copy_tree_with_hidden(src, tgt)
    src = 'nist-content/nist.gov/SP800-53/rev4/json'
    tgt = workspace / src
    test_utils.copy_tree_with_hidden(src, tgt)
    cwd = os.getcwd()
    try:
        os.chdir(workspace)
        command = 'trestle init -v'
        test_utils.execute_command_and_assert(command, 0, monkeypatch)
    finally:
        os.chdir(cwd)


def test_execute_validate_controls(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test execute validate_controls."""
    workspace = tmp_path / 'trestle.workspace'
    _setup_workspace(workspace, monkeypatch)
    cwd = os.getcwd()
    try:
        os.chdir(workspace)
        _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        section['validate-controls'] = 'warn'
        retval = tgt.execute()
    finally:
        os.chdir(cwd)
    assert retval == TaskOutcome.SUCCESS


def test_execute_validate_controls_nist(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test execute validate_controls nist."""
    workspace = tmp_path / 'trestle.workspace'
    _setup_workspace(workspace, monkeypatch)
    cwd = os.getcwd()
    try:
        os.chdir(workspace)
        _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
        # replace resolved profile (catalog)
        rows = _get_rows('tests/data/csv/ocp4-user.v2.csv')
        for i, row in enumerate(rows):
            if i == 0:
                assert row[3] == 'Profile_Source'
                continue
            elif i == 1:
                continue
            assert row[3] == 'tests/data/ocp4-cis/catalog.json'
            row[3] = 'nist-content/nist.gov/SP800-53/rev4/json/NIST_SP-800-53_rev4_catalog.json'
        # test
        with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
            mock_csv_reader.return_value = rows
            section['validate-controls'] = 'warn'
            tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
            retval = tgt.execute()
    finally:
        os.chdir(cwd)
    assert retval == TaskOutcome.SUCCESS


def test_execute_control_invalid(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test execute control invalid (not in catalog)."""
    workspace = tmp_path / 'trestle.workspace'
    _setup_workspace(workspace, monkeypatch)
    cwd = os.getcwd()
    try:
        os.chdir(workspace)
        _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
        # inject error
        rows = _get_rows('tests/data/csv/ocp4-user.v2.csv')
        row = rows[2]
        assert row[6] == 'CIS-1.2.1'
        row[6] = 'CIS-9.9.9'
        # test
        with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
            mock_csv_reader.return_value = rows
            section['validate-controls'] = 'warn'
            tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
            retval = tgt.execute()
    finally:
        os.chdir(cwd)
    assert retval == TaskOutcome.SUCCESS


def test_execute_control_invalid_fail(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test execute control invalid (not in catalog)."""
    workspace = tmp_path / 'trestle.workspace'
    _setup_workspace(workspace, monkeypatch)
    cwd = os.getcwd()
    try:
        os.chdir(workspace)
        _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
        # inject error
        rows = _get_rows('tests/data/csv/ocp4-user.v2.csv')
        row = rows[2]
        assert row[6] == 'CIS-1.2.1'
        row[6] = 'CIS-9.9.9'
        # test
        with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
            mock_csv_reader.return_value = rows
            section['validate-controls'] = 'on'
            tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
            retval = tgt.execute()
    finally:
        os.chdir(cwd)
    assert retval == TaskOutcome.FAILURE


def test_execute_no_overwrite(tmp_path: pathlib.Path) -> None:
    """Test execute no overwrite."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    section['output-overwrite'] = 'false'
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_execute_verbose(tmp_path: pathlib.Path) -> None:
    """Test execute verbose."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    section['quiet'] = 'False'
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_ocp(tmp_path)


def test_execute_missing_heading(tmp_path: pathlib.Path) -> None:
    """Test execute missing heading."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    # inject error
    rows = _get_rows('tests/data/csv/ocp4-user.v2.csv')
    row = rows[0]
    assert row[2] == 'Rule_Description'
    row[2] = 'foobar'
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_execute_missing_value(tmp_path: pathlib.Path) -> None:
    """Test execute missing value."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    # inject error
    rows = _get_rows('tests/data/csv/ocp4-user.v2.csv')
    row = rows[2]
    assert row[2] == 'Ensure that the --anonymous-auth argument is set to false'
    row[2] = ''
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_execute_missing_rule_id(tmp_path: pathlib.Path) -> None:
    """Test execute missing rule id."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    # inject error
    rows = _get_rows('tests/data/csv/ocp4-user.v2.csv')
    row = rows[2]
    assert row[1] == 'xccdf_org.ssgproject.content_rule_api_server_anonymous_auth'
    row[1] = ''
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_execute_missing_control_id_list(tmp_path: pathlib.Path) -> None:
    """Test execute missing control id list."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd.config')
    # inject error
    rows = _get_rows('tests/data/csv/ocp4-user.v2.csv')
    row = rows[2]
    assert row[6] == 'CIS-1.2.1'
    row[6] = ''
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 423


def test_execute_bp_sample(tmp_path: pathlib.Path) -> None:
    """Test execute bp sample."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_bp(tmp_path)


def test_execute_bp3_sample(tmp_path: pathlib.Path) -> None:
    """Test execute bp3 sample."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['csv-file'] = 'tests/data/csv/bp.sample.v3.csv'
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 59
    assert len(component.control_implementations) == 1
    ci = component.control_implementations[0]
    assert len(ci.set_parameters) == 4
    assert len(ci.set_parameters[0].values) == 1
    assert len(ci.set_parameters[1].values) == 3
    assert ci.set_parameters[1].values[0] == 'x'
    assert ci.set_parameters[1].values[1] == 'y'
    assert ci.set_parameters[1].values[2] == 'z'
    assert len(ci.set_parameters[2].values) == 3
    assert len(ci.set_parameters[3].values) == 3
    assert ci.set_parameters[3].values[0] == '3'
    assert ci.set_parameters[3].values[1] == '4'
    assert ci.set_parameters[3].values[2] == '5'


def test_execute_bp4_sample(tmp_path: pathlib.Path) -> None:
    """Test execute bp4 sample."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['csv-file'] = 'tests/data/csv/bp.sample.v4.csv'
    # perform transformation from csv to OSCAL json
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 68
    assert len(component.control_implementations) == 1
    # props
    index = 0
    assert component.props[index].name == 'Rule_Id'
    assert component.props[index].value == 'account_owner_authorized_ip_range_configured'
    assert component.props[index].remarks == 'rule_set_00'
    index = 7
    assert component.props[index].name == 'Parameter_Id'
    assert component.props[index].value == 'allowed_admins_per_account'
    assert component.props[index].remarks == 'rule_set_01'
    index = 8
    assert component.props[index].name == 'Parameter_Description'
    assert component.props[index].value == 'Maximum allowed administrators per'
    assert component.props[index].remarks == 'rule_set_01'
    index = 9
    assert component.props[index].name == 'Parameter_Value_Alternatives'
    assert component.props[index].value == '10'
    assert component.props[index].remarks == 'rule_set_01'
    index = 10
    assert component.props[index].name == 'Parameter_Id2'
    assert component.props[index].value == 'allowed_admins_per_account2'
    assert component.props[index].remarks == 'rule_set_01'
    index = 11
    assert component.props[index].name == 'Parameter_Description2'
    assert component.props[index].value == 'Maximum allowed administrators per2'
    assert component.props[index].remarks == 'rule_set_01'
    index = 12
    assert component.props[index].name == 'Parameter_Value_Alternatives2'
    assert component.props[index].value == '20'
    assert component.props[index].remarks == 'rule_set_01'
    index = 13
    assert component.props[index].name == 'Check_Id'
    assert component.props[index].value == 'iam_admin_role_users_per_account_maxcount'
    assert component.props[index].remarks == 'rule_set_01'
    # control implementations
    ci = component.control_implementations[0]
    assert len(ci.set_parameters) == 8
    index = 0
    assert ci.set_parameters[index].param_id == 'allowed_admins_per_account'
    assert len(ci.set_parameters[index].values) == 1
    assert ci.set_parameters[index].values[0] == '10'
    index = 1
    assert ci.set_parameters[index].param_id == 'allowed_admins_per_account2'
    assert len(ci.set_parameters[index].values) == 1
    assert ci.set_parameters[index].values[0] == '20'
    index = 2
    assert ci.set_parameters[index].param_id == 'api_keys_rotated_days'
    assert len(ci.set_parameters[index].values) == 3
    assert ci.set_parameters[index].values[0] == 'x'
    assert ci.set_parameters[index].values[1] == 'y'
    assert ci.set_parameters[index].values[2] == 'z'
    index = 3
    assert ci.set_parameters[index].param_id == 'api_keys_rotated_days2'
    assert len(ci.set_parameters[index].values) == 3
    assert ci.set_parameters[index].values[0] == 'r'
    assert ci.set_parameters[index].values[1] == 's'
    assert ci.set_parameters[index].values[2] == 't'


def test_execute_bp4_modify_additional_param_set(tmp_path: pathlib.Path) -> None:
    """Test execute bp4 modify additional param set."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['csv-file'] = 'tests/data/csv/bp.sample.v4.csv'
    # modify additional param set
    rows = _get_rows('tests/data/csv/bp.sample.v4.csv')
    row = rows[3]
    assert row[17] == 'allowed_admins_per_account2'
    assert row[19] == '20'
    assert row[20] == '20'
    row[19] = '50'
    row[20] = '45 50 55'
    # perform transformation from csv to OSCAL json
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 68
    assert len(component.control_implementations) == 1
    # props
    index = 0
    assert component.props[index].name == 'Rule_Id'
    assert component.props[index].value == 'account_owner_authorized_ip_range_configured'
    assert component.props[index].remarks == 'rule_set_00'
    index = 12
    assert component.props[index].name == 'Parameter_Value_Alternatives2'
    assert component.props[index].value == '45 50 55'
    assert component.props[index].remarks == 'rule_set_01'
    # control implementations
    ci = component.control_implementations[0]
    assert len(ci.set_parameters) == 8
    index = 0
    assert ci.set_parameters[index].param_id == 'allowed_admins_per_account'
    assert len(ci.set_parameters[index].values) == 1
    assert ci.set_parameters[index].values[0] == '10'
    index = 1
    assert ci.set_parameters[index].param_id == 'allowed_admins_per_account2'
    assert len(ci.set_parameters[index].values) == 1
    assert ci.set_parameters[index].values[0] == '50'


def test_execute_bp4_delete_additional_param_set(tmp_path: pathlib.Path) -> None:
    """Test execute bp4 delete additional param set."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['csv-file'] = 'tests/data/csv/bp.sample.v4.csv'
    # delete additional param set
    rows = _get_rows('tests/data/csv/bp.sample.v4.csv')
    row = rows[3]
    assert row[17] == 'allowed_admins_per_account2'
    assert row[19] == '20'
    assert row[20] == '20'
    row[17] = ''
    row[18] = ''
    row[19] = ''
    row[20] = ''
    # perform transformation from csv to OSCAL json
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 65
    assert len(component.control_implementations) == 1
    # control implementations
    ci = component.control_implementations[0]
    assert len(ci.set_parameters) == 7


def test_execute_bp_cd(tmp_path: pathlib.Path) -> None:
    """Test execute bp cd."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_bp(tmp_path)


def test_execute_bp_cd_missing(tmp_path: pathlib.Path) -> None:
    """Test execute bp cd missing."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/foobar/component-definition.json'
    tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_execute_duplicate_rule(tmp_path: pathlib.Path) -> None:
    """Test execute duplicate rule."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # duplicate rule
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    rows.append(rows[2])
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_execute_delete_rule(tmp_path: pathlib.Path) -> None:
    """Test execute delete rule."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # delete rule
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    rows.pop(1)
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 57
    assert component.props[0].name == 'Rule_Id'
    assert component.props[0].ns == 'http://abc.github.io/compliance-trestle/schemas/oscal/cd'
    assert component.props[0].value == 'iam_admin_role_users_per_account_maxcount'
    assert component.props[0].class_ == 'scc_class'
    assert component.props[0].remarks == 'rule_set_01'


def test_execute_delete_all_rules_with_params(tmp_path: pathlib.Path) -> None:
    """Test execute delete all rules with params."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # delete all rules with params
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    rows.pop(11)
    rows.pop(8)
    rows.pop(6)
    rows.pop(3)
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 30
    for prop in component.props:
        if prop.name == 'Parameter_Id':
            assert prop.value != 'allowed_admins_per_account'


def test_execute_delete_rule_with_params(tmp_path: pathlib.Path) -> None:
    """Test execute delete rule with params."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # delete rule with params
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    rows.pop(3)
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 54
    for prop in component.props:
        if prop.name == 'Parameter_Id':
            assert prop.value != 'allowed_admins_per_account'


def test_execute_column_bad_data(tmp_path: pathlib.Path) -> None:
    """Test execute column bad data."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # add rule
    rows = _get_rows('tests/data/csv/soc2.sample.v1.csv')
    rows[3][2] = '```'
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_execute_column_ignore(tmp_path: pathlib.Path) -> None:
    """Test execute column ignore."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # add rule
    rows = _get_rows('tests/data/csv/soc2.sample.v2.csv')
    rows[3][2] = '```'
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS


def test_execute_add_rule(tmp_path: pathlib.Path) -> None:
    """Test execute add rule."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # add rule
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    row = [
        'add-reference-id',
        'add-rule-id',
        'add-rule-description',
        'add-check-id',
        'add-check-description',
        'add-fetcher',
        'add-fetcher-description',
        'https://abc.com/add-profile-reference-url',
        'add-profile-description',
        'Service',
        'add-control-id-list',
        'IAM',
        'IAM',
        'add-parameter-id',
        'add-parameter-description',
        'add-parameter-value-default',
        'add-parameter-value-alternatives',
        'https://abc.com/add-namespace'
    ]
    rows.append(row)
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    pc = 72
    assert len(component.props) == 72
    assert component.props[pc - 1].name == 'Fetcher_Description'
    assert component.props[pc - 1].ns == 'https://abc.com/add-namespace'
    assert component.props[pc - 1].value == 'add-fetcher-description'
    assert component.props[pc - 1].remarks == 'rule_set_10'
    assert component.props[pc - 2].name == 'Fetcher'
    assert component.props[pc - 2].ns == 'https://abc.com/add-namespace'
    assert component.props[pc - 2].value == 'add-fetcher'
    assert component.props[pc - 2].remarks == 'rule_set_10'
    assert component.props[pc - 3].name == 'Reference_Id'
    assert component.props[pc - 3].ns == 'https://abc.com/add-namespace'
    assert component.props[pc - 3].value == 'add-reference-id'
    assert component.props[pc - 3].remarks == 'rule_set_10'
    assert component.props[pc - 4].name == 'Check_Description'
    assert component.props[pc - 4].ns == 'https://abc.com/add-namespace'
    assert component.props[pc - 4].value == 'add-check-description'
    assert component.props[pc - 4].remarks == 'rule_set_10'
    assert component.props[pc - 5].name == 'Check_Id'
    assert component.props[pc - 5].ns == 'https://abc.com/add-namespace'
    assert component.props[pc - 5].value == 'add-check-id'
    assert component.props[pc - 5].remarks == 'rule_set_10'
    assert component.props[pc - 6].name == 'Parameter_Value_Alternatives'
    assert component.props[pc - 6].ns == 'https://abc.com/add-namespace'
    assert component.props[pc - 6].value == 'add-parameter-value-alternatives'
    assert component.props[pc - 6].remarks == 'rule_set_10'
    assert component.props[pc - 7].name == 'Parameter_Description'
    assert component.props[pc - 7].ns == 'https://abc.com/add-namespace'
    assert component.props[pc - 7].value == 'add-parameter-description'
    assert component.props[pc - 7].remarks == 'rule_set_10'
    assert component.props[pc - 8].name == 'Parameter_Id'
    assert component.props[pc - 8].ns == 'https://abc.com/add-namespace'
    assert component.props[pc - 8].value == 'add-parameter-id'
    assert component.props[pc - 8].remarks == 'rule_set_10'
    assert component.props[pc - 9].name == 'Rule_Description'
    assert component.props[pc - 9].ns == 'https://abc.com/add-namespace'
    assert component.props[pc - 9].value == 'add-rule-description'
    assert component.props[pc - 9].remarks == 'rule_set_10'
    assert component.props[pc - 10].name == 'Rule_Id'
    assert component.props[pc - 10].ns == 'https://abc.com/add-namespace'
    assert component.props[pc - 10].value == 'add-rule-id'
    assert component.props[pc - 10].remarks == 'rule_set_10'
    assert len(component.control_implementations) == 2
    assert len(component.control_implementations[0].set_parameters) == 4
    assert len(component.control_implementations[1].set_parameters) == 1


def test_execute_param_duplicate_value(tmp_path: pathlib.Path) -> None:
    """Test execute param duplicate default value."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    # duplicate default param default value
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    row = rows[3]
    assert row[13] == 'allowed_admins_per_account'
    assert row[15] == '10'
    row = rows[2]
    row[13] = 'allowed_admins_per_account'
    row[15] = '10'
    rows[2] = row
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    row[15] = '11'
    rows[2] = row
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_execute_missing_param_default_value(tmp_path: pathlib.Path) -> None:
    """Test execute missing param default_value."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    # delete default param default value
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    row = rows[3]
    assert row[13] == 'allowed_admins_per_account'
    assert row[15] == '10'
    row[15] = ''
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS


def test_execute_change_param_default_value(tmp_path: pathlib.Path) -> None:
    """Test execute change param default_value."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # change default param default value
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    row = rows[3]
    assert row[13] == 'allowed_admins_per_account'
    assert row[15] == '10'
    assert row[16] == '10'
    row[15] = '20'
    row[16] = '10 20'
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 62
    assert len(component.control_implementations) == 1
    set_parameters = component.control_implementations[0].set_parameters
    assert len(set_parameters) == 4
    assert len(set_parameters[0].values) == 1
    assert set_parameters[0].values[0] == '20'


def test_execute_delete_param(tmp_path: pathlib.Path) -> None:
    """Test execute delete param."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # delete param
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    row = rows[3]
    assert row[13] == 'allowed_admins_per_account'
    row[13] = ''
    row[14] = ''
    row[15] = ''
    row[16] = ''
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 59
    assert component.props[22].name == 'Parameter_Id'
    assert component.props[22].ns == 'http://abc.github.io/compliance-trestle/schemas/oscal/cd'
    assert component.props[22].value == 'api_keys_rotated_days'
    assert component.props[22].class_ == 'scc_class'
    assert component.props[22].remarks == 'rule_set_04'


def test_execute_delete_params(tmp_path: pathlib.Path) -> None:
    """Test execute delete params."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # delete params
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    for i in [3, 6, 8, 11]:
        for j in [13, 14, 15, 16]:
            rows[i][j] = ''
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 50
    for prop in component.props:
        assert prop.name != 'Parameter_Id'


def test_execute_add_param(tmp_path: pathlib.Path) -> None:
    """Test execute add param."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # add param
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    row = rows[2]
    assert row[13] == ''
    row[13] = 'add-parameter-id'
    row[14] = 'add-parameter-description'
    row[15] = 'add-parameter-default-value'
    row[16] = 'add-parameter-value-alternatives'
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert component.props[2].name == 'Parameter_Id'
    assert component.props[2].value == 'add-parameter-id'
    assert component.props[2].remarks == 'rule_set_00'
    assert component.props[3].name == 'Parameter_Description'
    assert component.props[3].value == 'add-parameter-description'
    assert component.props[3].remarks == 'rule_set_00'
    assert component.props[4].name == 'Parameter_Value_Alternatives'
    assert component.props[4].value == 'add-parameter-value-alternatives'
    assert component.props[4].remarks == 'rule_set_00'
    assert component.props[4].remarks == 'rule_set_00'
    set_parameters = component.control_implementations[0].set_parameters
    assert len(set_parameters) == 5
    assert set_parameters[4].param_id == 'add-parameter-id'
    assert len(set_parameters[4].values) == 1
    assert set_parameters[4].values[0] == 'add-parameter-default-value'


def test_execute_delete_all_control_id_list(tmp_path: pathlib.Path) -> None:
    """Test execute delete all control id list."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # delete all control lists
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    for i in range(1, len(rows)):
        rows[i][10] = ''
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    if cd.components is not None:
        assert len(cd.components) == 0


def test_execute_delete_control_id(tmp_path: pathlib.Path) -> None:
    """Test execute delete control id."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # delete control id
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    assert rows[2][10] == 'sc-7_smt.a sc-7_smt.b sc-7.3 sc-7.4_smt.a sc-7.5 ia-3'
    rows[2][10] = rows[2][10].replace(' ia-3', '')
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    control_implementation = component.control_implementations[0]
    implemented_requirements = control_implementation.implemented_requirements
    assert len(implemented_requirements) == 20
    assert implemented_requirements[3].control_id == 'sc-7.5'
    assert implemented_requirements[4].control_id == 'ac-6'


def test_execute_delete_control_id_multi(tmp_path: pathlib.Path) -> None:
    """Test execute delete control id multi."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # delete control id multi
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    for i in range(1, len(rows)):
        rows[i][10] = rows[i][10].replace('ac-6', '')
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    control_implementation = component.control_implementations[0]
    implemented_requirements = control_implementation.implemented_requirements
    assert len(implemented_requirements) == 20
    for implemented_requirement in implemented_requirements:
        assert implemented_requirement.control_id != 'ac-6'


def test_execute_delete_control_id_smt(tmp_path: pathlib.Path) -> None:
    """Test execute delete control id smt."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # delete control id smt
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    assert rows[4][10] == 'ac-3 ac-4 ac-6 sc-7_smt.a sc-7_smt.b sc-7.4_smt.a ac-14_smt.a cm-7_smt.a cm-7_smt.b'
    rows[4][10] = rows[4][10].replace(' cm-7_smt.b', '')
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    control_implementation = component.control_implementations[0]
    implemented_requirements = control_implementation.implemented_requirements
    assert len(implemented_requirements) == 21
    assert implemented_requirements[10].control_id == 'cm-7'
    statements = implemented_requirements[10].statements
    assert len(statements) == 1
    assert statements[0].statement_id == 'cm-7_smt.a'


def test_execute_add_control_id(tmp_path: pathlib.Path) -> None:
    """Test execute add control id."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # add control id
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    assert rows[2][10] == 'sc-7_smt.a sc-7_smt.b sc-7.3 sc-7.4_smt.a sc-7.5 ia-3'
    rows[2][10] = rows[2][10] + ' ld-0'
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    control_implementation = component.control_implementations[0]
    implemented_requirements = control_implementation.implemented_requirements
    assert len(implemented_requirements) == 22
    assert implemented_requirements[21].control_id == 'ld-0'


def test_execute_add_control_id_smt(tmp_path: pathlib.Path) -> None:
    """Test execute add control mapping smt."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # add control mapping smt
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    assert rows[2][10] == 'sc-7_smt.a sc-7_smt.b sc-7.3 sc-7.4_smt.a sc-7.5 ia-3'
    rows[2][10] = rows[2][10] + ' ld-0_smt.a'
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    control_implementation = component.control_implementations[0]
    implemented_requirements = control_implementation.implemented_requirements
    assert len(implemented_requirements) == 22
    assert implemented_requirements[21].control_id == 'ld-0'
    assert len(implemented_requirements[21].statements) == 1
    statement = implemented_requirements[21].statements[0]
    assert statement.statement_id == 'ld-0_smt.a'
    assert len(statement.props) == 1
    assert statement.props[0].name == 'Rule_Id'
    assert statement.props[0].value == 'account_owner_authorized_ip_range_configured'


def test_execute_delete_property(tmp_path: pathlib.Path) -> None:
    """Test execute delete property."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # delete property
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    assert rows[2][3] == 'account_owner_authorized_ip_range_configured'
    assert rows[2][4] == 'Check whether authorized IP ranges are configured by the account owner'
    rows[2][3] = ''
    rows[2][4] = ''
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 60
    assert component.props[0].name == 'Rule_Id'
    assert component.props[1].name == 'Rule_Description'
    assert component.props[2].name == 'Reference_Id'


def test_execute_add_property(tmp_path: pathlib.Path) -> None:
    """Test execute add property."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # add property
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    assert rows[2][5] == ''
    assert rows[2][6] == ''
    rows[2][5] = 'add-fetcher'
    rows[2][6] = 'add-fetcher-description'
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 64
    assert component.props[0].name == 'Rule_Id'
    assert component.props[1].name == 'Rule_Description'
    assert component.props[2].name == 'Check_Id'
    assert component.props[3].name == 'Check_Description'
    assert component.props[4].name == 'Reference_Id'
    assert component.props[5].name == 'Fetcher'
    assert component.props[5].value == 'add-fetcher'
    assert component.props[6].name == 'Fetcher_Description'
    assert component.props[6].value == 'add-fetcher-description'


def test_execute_add_user_property(tmp_path: pathlib.Path) -> None:
    """Test execute add user property."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # add user property
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    rows[0].append('New_Column_Name')
    rows[1].append('New_Column_Name column description')
    for i in range(2, len(rows)):
        rows[i].append(f'new-column-value-{i}')
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[0]
    assert len(component.props) == 71
    assert component.props[5].name == 'New_Column_Name'
    assert component.props[5].value == 'new-column-value-2'


def test_execute_validation(tmp_path: pathlib.Path) -> None:
    """Test execute validation."""
    _, section = _get_config_section_init(tmp_path, 'test-csv-to-oscal-cd-bp.config')
    section['component-definition'] = 'tests/data/csv/component-definitions/bp/component-definition.json'
    # set validation component type
    rows = _get_rows('tests/data/csv/bp.sample.v2.csv')
    row = [
        'validation-reference-id',
        'validation-rule-id',
        'validation-rule-description',
        'validation-check-id',
        'validation-check-description',
        'validation-fetcher',
        'validation-fetcher-description',
        'https://abc.com/validation-profile-reference-url',
        'validation-profile-description',
        'Validation',
        'validation-control-id-list',
        'IAM',
        'IAM',
        'validation-parameter-id',
        'validation-parameter-description',
        'validation-parameter-value-default',
        'validation-parameter-value-alternatives',
        'https://abc.com/validation-namespace'
    ]
    rows.append(row)
    with mock.patch('trestle.tasks.csv_to_oscal_cd.csv.reader') as mock_csv_reader:
        mock_csv_reader.return_value = rows
        tgt = csv_to_oscal_cd.CsvToOscalComponentDefinition(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
    # read component-definition
    fp = pathlib.Path(tmp_path) / 'component-definition.json'
    cd = ComponentDefinition.oscal_read(fp)
    # spot check
    component = cd.components[1]
    assert component.type == 'Validation'
    assert component.title == 'IAM'
    assert component.description == 'IAM'
    assert len(component.props) == 6
    assert component.props[0].name == 'Rule_Id'
    assert component.props[0].value == 'validation-rule-id'
    assert component.props[0].class_ == 'scc_class'
    assert component.props[1].name == 'Check_Id'
    assert component.props[1].value == 'validation-check-id'
    assert component.props[2].name == 'Check_Description'
    assert component.props[2].value == 'validation-check-description'
    assert len(component.control_implementations) == 0


def test_row_property_builder(tmp_path):
    """Test row property builder."""
    # valid
    prop = csv_to_oscal_cd.row_property_builder(
        row=0,
        name='name',
        value='value',
        ns='https://www.ibm.com',
        class_='class',
        remarks='remarks',
    )
    assert prop
    # missing name
    try:
        prop = csv_to_oscal_cd.row_property_builder(
            row=0,
            name=None,
            value='value',
            ns='https://www.ibm.com',
            class_='class',
            remarks='remarks',
        )
        raise AssertionError('missing name OK?')
    except Exception:
        assert prop
    # missing value
    try:
        prop = csv_to_oscal_cd.row_property_builder(
            row=0,
            name='name',
            value=None,
            ns='https://www.ibm.com',
            class_='class',
            remarks='remarks',
        )
        raise AssertionError('missing value OK?')
    except Exception:
        assert prop
    # invalid ns
    try:
        prop = csv_to_oscal_cd.row_property_builder(
            row=0,
            name='name',
            value='value',
            ns='foobar',
            class_='class',
            remarks='remarks',
        )
        raise AssertionError('invalid ns OK?')
    except Exception:
        assert prop
    # invalid class
    try:
        prop = csv_to_oscal_cd.row_property_builder(
            row=0,
            name='name',
            value='value',
            ns='https://www.ibm.com',
            class_='\n',
            remarks='remarks',
        )
        raise AssertionError('invalid class OK?')
    except Exception:
        assert prop



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for cli module command validate."""
import argparse
import pathlib
import shutil
import sys
from uuid import uuid4

from _pytest.monkeypatch import MonkeyPatch

import pytest

from tests import test_utils
from tests.test_utils import setup_for_ssp

import trestle.common.const as const
import trestle.core.generators as gens
import trestle.oscal.assessment_plan as ap
import trestle.oscal.profile as prof
import trestle.oscal.ssp as ossp
from trestle import cli
from trestle.cli import Trestle
from trestle.common.model_utils import ModelUtils
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.commands.split import SplitCmd
from trestle.core.generators import generate_sample_model
from trestle.core.models.file_content_type import FileContentType
from trestle.core.validator import Validator
from trestle.core.validator_factory import validator_factory
from trestle.oscal.catalog import Catalog
from trestle.oscal.common import ResponsibleParty, Role
from trestle.oscal.component import ComponentDefinition, ControlImplementation

test_data_dir = pathlib.Path('tests/data').resolve()

md_path = 'md_comp'


@pytest.mark.parametrize(
    'name, mode, parent',
    [
        ('my_test_model', '-f', False), ('my_test_model', '-n', False), ('my_test_model', '-f', True),
        ('my_test_model', '-t', False), ('my_test_model', '-a', False), ('my_test_model', '-x', False)
    ]
)
def test_validation_happy(name, mode, parent, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test successful validation runs."""
    (tmp_trestle_dir / test_utils.CATALOGS_DIR / 'my_test_model').mkdir(exist_ok=True, parents=True)
    (tmp_trestle_dir / test_utils.CATALOGS_DIR / 'my_test_model2').mkdir(exist_ok=True, parents=True)
    shutil.copyfile(
        test_data_dir / 'json/minimal_catalog.json',
        tmp_trestle_dir / test_utils.CATALOGS_DIR / 'my_test_model/catalog.json'
    )
    shutil.copyfile(
        test_data_dir / 'json/minimal_catalog.json',
        tmp_trestle_dir / test_utils.CATALOGS_DIR / 'my_test_model2/catalog.json'
    )

    model_def_file = tmp_trestle_dir / test_utils.CATALOGS_DIR / ('my_test_model/catalog.json')

    if mode == '-f':
        if not parent:
            testcmd = f'trestle validate {mode} {model_def_file}'
        else:
            testcmd = f'trestle validate {mode} {model_def_file.parent}'
    elif mode == '-n':
        testcmd = f'trestle validate -t catalog -n {name}'
    elif mode == '-x':
        testcmd = f'trestle validate -t catalog -n {name}'
    else:
        testcmd = 'trestle validate -a'

    monkeypatch.setattr(sys, 'argv', testcmd.split())
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        cli.run()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0


@pytest.mark.parametrize(
    'name, mode, parent, status',
    [
        ('my_test_model', '-f', False, 1), ('my_test_model', '-n', False, 4), ('my_test_model', '-f', True, 1),
        ('my_test_model', '-t', False, 1), ('my_test_model', '-a', False, 1), ('foo', '-n', False, 4),
        ('my_test_model', '-x', False, 4)
    ]
)
def test_validation_unhappy(
    name, mode, parent, status, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test failure modes of validation."""
    (tmp_trestle_dir / test_utils.CATALOGS_DIR / 'my_test_model').mkdir(exist_ok=True, parents=True)
    (tmp_trestle_dir / test_utils.CATALOGS_DIR / 'my_test_model2').mkdir(exist_ok=True, parents=True)
    shutil.copyfile(
        test_data_dir / 'json/minimal_catalog_bad_oscal_version.json',
        tmp_trestle_dir / test_utils.CATALOGS_DIR / 'my_test_model/catalog.json'
    )
    shutil.copyfile(
        test_data_dir / 'json/minimal_catalog.json',
        tmp_trestle_dir / test_utils.CATALOGS_DIR / 'my_test_model2/catalog.json'
    )

    model_def_file = tmp_trestle_dir / test_utils.CATALOGS_DIR / ('my_test_model/catalog.json')

    if mode == '-f':
        if not parent:
            testcmd = f'trestle validate {mode} {model_def_file}'
        else:
            testcmd = f'trestle validate {mode} {model_def_file.parent}'
    elif mode == '-n':
        testcmd = f'trestle validate -t catalog -n {name}'
    elif mode == '-x':
        testcmd = f'trestle validate -t catalog -n {name}'
    else:
        testcmd = 'trestle validate -a'

    monkeypatch.setattr(sys, 'argv', testcmd.split())
    rc = Trestle().run()
    assert rc == status


@pytest.mark.parametrize(
    'name, mode, parent, test_id, code',
    [
        ('my_ap', '-f', False, 'id1', 0), ('my_ap', '-n', False, 'id1', 0), ('my_ap', '-f', True, 'id1', 0),
        ('my_ap', '-t', False, 'id1', 0), ('my_ap', '-a', False, 'id1', 0), ('my_ap', '-f', False, 'foo', 4),
        ('my_ap', '-n', False, 'foo', 4), ('my_ap', '-f', True, 'foo', 4), ('my_ap', '-t', False, 'foo', 4),
        ('my_ap', '-a', False, 'foo', 4), ('foo', '-n', False, 'id1', 4)
    ]
)
def test_role_refs_validator(
    name, mode, parent, test_id, code, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test validation of roles and references to them in responsible-parties."""
    (tmp_trestle_dir / 'assessment-plans/my_ap').mkdir(exist_ok=True, parents=True)
    roles = [Role(id='id1', title='title1'), Role(id='id2', title='title2'), Role(id='id3', title='title3')]
    party1 = ResponsibleParty(role_id=test_id, party_uuids=[str(uuid4())])
    party2 = ResponsibleParty(role_id='id2', party_uuids=[str(uuid4())])
    responsible_parties = [party1, party2]
    ap_obj = generate_sample_model(ap.AssessmentPlan)
    ap_obj.metadata.roles = roles
    ap_obj.metadata.responsible_parties = responsible_parties
    ap_path = tmp_trestle_dir / 'assessment-plans/my_ap/assessment-plan.json'
    ap_obj.oscal_write(ap_path)

    if mode == '-f':
        if not parent:
            testcmd = f'trestle validate {mode} {ap_path}'
        else:
            testcmd = f'trestle validate {mode} {ap_path.parent}'
    elif mode == '-n':
        testcmd = f'trestle validate -t assessment-plan -n {name}'
    elif mode == '-t':
        testcmd = 'trestle validate -t assessment-plan'
    else:
        testcmd = 'trestle validate -a'

    monkeypatch.setattr(sys, 'argv', testcmd.split())
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        cli.run()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == code


def test_oscal_version_validator(
    tmp_trestle_dir: pathlib.Path, sample_catalog_minimal: Catalog, monkeypatch: MonkeyPatch
) -> None:
    """
    Test oscal version validator.

    There is no actual validator for oscal version but it happens in base_model when a file is loaded.
    """
    mycat_dir = tmp_trestle_dir / 'catalogs/mycat'
    mycat_dir.mkdir()
    mycat_path = mycat_dir / 'catalog.json'
    sample_catalog_minimal.oscal_write(mycat_path)
    testcmd = 'trestle validate -t catalog'
    monkeypatch.setattr(sys, 'argv', testcmd.split())
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        cli.run()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == CmdReturnCodes.SUCCESS.value


def test_oscal_version_incorrect_validator(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test validation fails for bad oscal version. Short pydantic message should be printed."""
    catalog_path = test_utils.JSON_TEST_DATA_PATH / 'minimal_catalog_bad_oscal_version.json'
    mycat_dir = tmp_trestle_dir / 'catalogs/mycat'
    mycat_dir.mkdir()
    mycat_path = mycat_dir / 'catalog.json'
    shutil.copyfile(catalog_path, mycat_path)
    testcmd = f'trestle validate -f {mycat_path}'
    monkeypatch.setattr(sys, 'argv', testcmd.split())
    rc = Trestle().run()
    assert rc == 1


def test_oscal_incorrect_fields_validator(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test validation fails for oscal with extra fields. Full pydantic message should be printed."""
    catalog_path = test_utils.JSON_TEST_DATA_PATH / 'minimal_catalog_extra_fields.json'
    mycat_dir = tmp_trestle_dir / 'catalogs/mycat'
    mycat_dir.mkdir()
    catalog = mycat_dir / 'catalog.json'
    catalog.touch()
    shutil.copyfile(catalog_path, catalog)
    testcmd = f'trestle validate -f {catalog}'
    monkeypatch.setattr(sys, 'argv', testcmd.split())
    rc = Trestle().run()
    assert rc == 1


def test_validate_direct(sample_catalog_minimal: Catalog, tmp_trestle_dir: pathlib.Path) -> None:
    """Test a validator by invoking it directly without CLI."""
    args = argparse.Namespace(mode=const.VAL_MODE_ALL, quiet=True)
    validator: Validator = validator_factory.get(args)
    assert validator.model_is_valid(sample_catalog_minimal, True, tmp_trestle_dir)


def test_validate_dup_uuids(
    sample_component_definition: ComponentDefinition, sample_trestle_profile: prof.Profile
) -> None:
    """Test validation of comp def with duplicate uuids."""
    args = argparse.Namespace(mode=const.VAL_MODE_ALL, quiet=True)
    validator = validator_factory.get(args)

    # confirm the comp_def has duplicate param_ids (REPLACE_ME)
    ci = sample_component_definition.components[0].control_implementations[0]
    assert ci.set_parameters[0].param_id == ci.implemented_requirements[0].set_parameters[0].param_id

    # confirm the comp_def is valid despite duplicate param_ids (this is allowed for compdef and ssp)
    assert validator.model_is_valid(sample_component_definition, False)

    # force two components to have same uuid and confirm invalid
    sample_component_definition.components[1].uuid = sample_component_definition.components[0].uuid
    assert not validator.model_is_valid(sample_component_definition, True, None)

    # restore uuid to unique value and confirm it is valid again
    sample_component_definition.components[1].uuid = str(uuid4())
    assert validator.model_is_valid(sample_component_definition, False, None)

    # add a control implementation to one of the components and confirm valid
    control_imp: ControlImplementation = generate_sample_model(ControlImplementation)
    sample_component_definition.components[1].control_implementations = [control_imp]
    assert validator.model_is_valid(sample_component_definition, True, None)

    # force the control implementation to have same uuid as the first component and confirm invalid
    sample_component_definition.components[1].control_implementations[0].uuid = sample_component_definition.components[
        0].uuid
    assert not validator.model_is_valid(sample_component_definition, True, None)

    assert validator.model_is_valid(sample_trestle_profile, True)

    # add extra set_param with duplicate param_id and confirm it is not valid
    # only profile has this additional test
    set_param = sample_trestle_profile.modify.set_parameters[0].copy()
    set_param.values = ['foo']
    sample_trestle_profile.modify.set_parameters.append(set_param)
    assert not validator.model_is_valid(sample_trestle_profile, True)


def test_validate_distributed(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Check that validate will run correctly when exploiting load distributed."""
    test_utils.ensure_trestle_config_dir(tmp_trestle_dir)
    # Clean up.
    test_data_source = testdata_dir / 'split_merge/step0-merged_catalog/catalogs'
    catalogs_dir = tmp_trestle_dir / 'catalogs'
    shutil.rmtree(catalogs_dir)
    shutil.copytree(test_data_source, catalogs_dir)

    args = argparse.Namespace(
        name='split',
        file='catalogs/mycatalog/catalog.json',
        verbose=1,
        element='catalog.groups.*.controls.*',
        trestle_root=tmp_trestle_dir
    )
    _ = SplitCmd()._run(args)
    test_args = 'trestle validate -a -v'.split(' ')
    monkeypatch.setattr(sys, 'argv', test_args)
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        cli.run()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0


def test_validate_catalog_params(sample_catalog_rich_controls: Catalog) -> None:
    """Test validation of unique param ids in catalog."""
    args = argparse.Namespace(mode=const.VAL_MODE_CATALOG)
    validator: Validator = validator_factory.get(args)
    assert validator.model_is_valid(sample_catalog_rich_controls, True, None)
    param_0_id = sample_catalog_rich_controls.groups[0].controls[0].params[0].id
    sample_catalog_rich_controls.groups[0].controls[0].params[1].id = param_0_id
    assert not validator.model_is_valid(sample_catalog_rich_controls, False, None)


def test_rule_param_values_validator_happy(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test validation of rules in SSP."""
    args = argparse.Namespace(mode=const.VAL_MODE_RULES)
    validator: Validator = validator_factory.get(args)
    prof_name = 'comp_prof'
    ssp_name = 'new_ssp'

    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    test_utils.gen_and_assemble_first_ssp(prof_name, ssp_name, gen_args, monkeypatch)

    new_ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    imp_reqs = new_ssp.control_implementation.implemented_requirements
    imp_req = next((i_req for i_req in imp_reqs if i_req.control_id == 'ac-3'), None)
    imp_req.by_components[0].set_parameters[1].values[0] = 'shared_param_1_aa_opt_1'
    ModelUtils.save_top_level_model(new_ssp, tmp_trestle_dir, ssp_name, FileContentType.JSON)

    assert validator.model_is_valid(new_ssp, True, tmp_trestle_dir)


def test_rule_param_values_validator_unhappy(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test validation of rules in SSP."""
    args = argparse.Namespace(mode=const.VAL_MODE_RULES)
    validator: Validator = validator_factory.get(args)
    prof_name = 'comp_prof'
    ssp_name = 'my_ssp'

    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    args_compdefs = gen_args.compdefs
    test_utils.gen_and_assemble_first_ssp(prof_name, ssp_name, gen_args, monkeypatch)

    orig_ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    # starts editing by grabbing by components from second imp req
    imp_reqs = orig_ssp.control_implementation.implemented_requirements
    imp_req = next((i_req for i_req in imp_reqs if i_req.control_id == 'ac-3'), None)
    by_components = imp_req.by_components
    # changes values for set parameter to test inequality
    by_components[0].set_parameters[1].values = ['shared_param_1_ab_opt_3']

    ModelUtils.save_top_level_model(orig_ssp, tmp_trestle_dir, ssp_name, FileContentType.JSON)
    assert not validator.model_is_valid(orig_ssp, True, tmp_trestle_dir)

    # test a by_component statement param value is added and
    # reassemble ssp
    ssp_assemble = f'trestle author ssp-assemble -m {ssp_name} -o {ssp_name} -cd {args_compdefs}'
    test_utils.execute_command_and_assert(ssp_assemble, 0, monkeypatch)
    # load new ssp
    new_ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)
    # grabs by component elements at statement level
    imp_reqs = new_ssp.control_implementation.implemented_requirements
    imp_req_ac = next((i_req for i_req in imp_reqs if i_req.control_id == 'ac-1'), None)
    by_components = imp_req_ac.statements[0].by_components
    by_components[1].set_parameters = []

    new_set_parameter_a = gens.generate_sample_model(ossp.SetParameter)
    new_set_parameter_a.values = ['shared_param_1_ab_opt_3']
    new_set_parameter_a.param_id = 'shared_param_1'
    # addes new value to current set parameter values
    # adds a new set parameter to set parameters array for current by component in statement
    by_components[1].set_parameters.append(new_set_parameter_a)
    imp_reqs = new_ssp.control_implementation.implemented_requirements
    imp_req_at = next((i_req for i_req in imp_reqs if i_req.control_id == 'at-1'), None)
    by_components = imp_req_at.statements[0].by_components
    by_components[0].set_parameters = []
    new_set_parameter_b = gens.generate_sample_model(ossp.SetParameter)
    new_set_parameter_b.values = ['shared_param_1_ab_opt_1']
    new_set_parameter_b.param_id = 'shared_param_1'
    by_components[0].set_parameters.append(new_set_parameter_b)

    ModelUtils.save_top_level_model(new_ssp, tmp_trestle_dir, ssp_name, FileContentType.JSON)
    assert not validator.model_is_valid(new_ssp, True, tmp_trestle_dir)
    # cleaning up the mess
    test_utils.gen_and_assemble_first_ssp(prof_name, ssp_name, gen_args, monkeypatch)


def test_validate_ssp_with_no_profile(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test validation of rules in SSP."""
    args = argparse.Namespace(mode=const.VAL_MODE_RULES)
    validator: Validator = validator_factory.get(args)
    prof_name = 'comp_prof'
    ssp_name = 'new_ssp'

    gen_args, _ = setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    test_utils.gen_and_assemble_first_ssp(prof_name, ssp_name, gen_args, monkeypatch)

    new_ssp, _ = ModelUtils.load_model_for_class(tmp_trestle_dir, ssp_name, ossp.SystemSecurityPlan)

    original_href = new_ssp.import_profile.href
    # sets href for given profile to empty to test if no given profile has been given to an ssp
    new_ssp.import_profile.href = ''
    ModelUtils.save_top_level_model(new_ssp, tmp_trestle_dir, ssp_name, FileContentType.JSON)

    assert not validator.model_is_valid(new_ssp, True, tmp_trestle_dir)

    validate_command = f'trestle validate -t system-security-plan -n {ssp_name}'
    test_utils.execute_command_and_assert(validate_command, 4, monkeypatch)

    new_ssp.import_profile.href = original_href
    ModelUtils.save_top_level_model(new_ssp, tmp_trestle_dir, ssp_name, FileContentType.JSON)



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Basic test script to test whether each reference in a json scema is referred to presuming there is one root model."""
import argparse
import json
import logging
import pathlib
import traceback
from typing import Any, Dict, List

from ilcli import Command

from trestle.core.commands.common.return_codes import CmdReturnCodes

logger = logging.getLogger(__name__)


class IntegrityCmd(Command):
    """Run arbitrary trestle tasks in a simple and extensible methodology."""

    name = 'schema_integrity'

    def _init_arguments(self) -> None:
        self.add_argument('directory', nargs='?', type=str, help='Path to the directory for oscal schema')

    def _run(self, args: argparse.Namespace) -> int:
        try:
            dir_of_schemas = pathlib.Path(args.directory)
            for path_of_schema in dir_of_schemas.glob('oscal_*_schema.json'):
                self.out(f'Examinining: {path_of_schema.name}')
                fp = path_of_schema.open('r', encoding='utf8')
                json_schema_raw = json.load(fp)
                fp.close()
                # get list of ID's
                # Cannot use this technique using latest OSCAL code drop
                # ids = list(set(self.recursive_ref('$id', json_schema_raw))) # noqa: E800
                # get list of ID's only from the top level elements in the definition structure.
                definitions = json_schema_raw['definitions']

                ids = []
                for _, value in definitions.items():
                    for key, vals in value.items():
                        if key == '$id':
                            ids.append(vals)

                refs = list(set(self.recursive_ref('$ref', json_schema_raw)))

                for id_ex in ids:
                    if (id_ex not in refs) and (id_ex[0] == '#'):
                        self.out(f'id {id_ex} not in use')

            return CmdReturnCodes.SUCCESS.value
        except Exception:
            self.err('Failure of script')
            self.err(traceback.format_exc())
            return CmdReturnCodes.COMMAND_ERROR.value

    def recursive_ref(self, ref_key: str, dict_of_interest: Dict[str, Any]) -> List[str]:
        """Return all values for keys where you look recursively.

        presume key of interest has a consistent type value
        """
        returner = []

        for key, value in dict_of_interest.items():
            if key == ref_key:
                returner.append(value)
            elif type(value) == dict:
                returner = returner + self.recursive_ref(ref_key, value)
            elif type(value) == list:
                for item in value:
                    if type(item) == dict:
                        returner = returner + self.recursive_ref(ref_key, item)
                    elif key == ref_key:
                        returner.append(value)
        return returner


if __name__ == '__main__':
    exit(IntegrityCmd().run())
